# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Core base backend specific functionality."""

from abc import ABC

from .actor import ActorClass
from .object_ref import ObjectRef
from .remote_function import RemoteFunction


class Backend(ABC):
    """An interface that represents the parent class for any backend class."""

    @staticmethod
    def make_remote_function(function, num_cpus, num_returns, resources):
        """
        Define a remote function.

        function : callable
            Function to be a remote function.
        num_cpus : int
            The number of CPUs to reserve for the remote function.
        num_returns : int
            The number of ``ObjectRef``-s returned by the remote function invocation.
        resources : dict
            Custom resources to reserve for the remote function.

        Returns
        -------
        RemoteFunction

        Notes
        -----
        The method of the child class for the concrete backend should return
        remote function object specific for the backend.
        """
        pass

    @staticmethod
    def make_actor(cls, num_cpus, resources):
        """
        Define an actor class.

        cls : object
            Class to be an actor class.
        num_cpus : int
            The number of CPUs to reserve for the lifetime of the actor.
        resources : dict
            Custom resources to reserve for the lifetime of the actor.

        Returns
        -------
        ActorClass

        Notes
        -----
        The method of the child class for the concrete backend should return
        actor class specific for the backend and arguments to be passed in the actor.
        """
        pass

    @staticmethod
    def get(object_refs):
        """
        Get a remote object or a list of remote objects
        from distributed memory.

        Parameters
        ----------
        object_refs : ObjectRef or list
            Object ref or a list of object refs to get data from.

        Returns
        -------
        object
            A Python object or a list of Python objects.

        Notes
        -----
        The method of the child class for the concrete backend should pass
        `object_refs` specific for the backend.
        """
        pass

    @staticmethod
    def put(data):
        """
        Put data into distributed memory.

        Parameters
        ----------
        data : object
            Data to be put.

        Returns
        -------
        ObjectRef
            ObjectRef matching to data.

        Notes
        -----
        The method of the child class for the concrete backend should return
        object ref specific for the backend.
        """
        pass

    @staticmethod
    def wait(object_refs, num_returns=1):
        """
        Wait until `object_refs` are finished.

        This method returns two lists. The first list consists of
        object ref-s that correspond to objects that completed computations.
        The second list corresponds to the rest of the object ref-s (which may or may not be ready).

        Parameters
        ----------
        object_refs : ObjectRef or list
            Object ref or list of object refs to be waited.
        num_returns : int, default: 1
            The number of object refs that should be returned as ready.

        Returns
        -------
        two lists
            List of object refs that are ready and list of the remaining object refs.

        Notes
        -----
        The method of the child class for the concrete backend should pass and return
        object ref-s specific for the backend.
        """
        pass

    @staticmethod
    def get_ip():
        """
        Get node IP address.

        Returns
        -------
        str
            Node IP address.
        """
        pass

    @staticmethod
    def num_cpus():
        """
        Get the number of CPUs used by the execution backend.

        Returns
        -------
        int
        """
        pass

    @staticmethod
    def cluster_resources():
        """
        Get resources of the cluster.

        Returns
        -------
        dict
            Dictionary with cluster nodes info in the form
            `{"node_ip0": {"CPU": x0}, "node_ip1": {"CPU": x1}, ...}`.
        """
        pass

    @staticmethod
    def shutdown():
        """Shutdown an execution backend."""
        pass

    @staticmethod
    def is_initialized():
        """
        Check if a unidist backend has already been initialized.

        Returns
        -------
        bool
            True or False.
        """
        pass


class BackendProxy(Backend):
    """
    A class which instance is a proxy object to dispatch operations to the concrete backend.

    Parameters
    ----------
    backend_cls : Backend
        Instance of the concrete backend class.
    """

    __instance = None

    def __init__(self, backend_cls):
        if self.__instance is None:
            self._backend_cls = backend_cls

    @classmethod
    def get_instance(cls, backend_cls=None):
        """
        Get instance of this class.

        Parameters
        ----------
        backend_cls : Backend, optional
            Instance of the concrete backend class.
        """
        if cls.__instance is None and backend_cls is not None:
            cls.__instance = BackendProxy(backend_cls)
        return cls.__instance

    def make_remote_function(self, function, num_cpus, num_returns, resources):
        """
        Define a remote function.

        function : callable
            Function to be a remote function.
        num_cpus : int
            The number of CPUs to reserve for the remote function.
        num_returns : int
            The number of ``ObjectRef``-s returned by the remote function invocation.
        resources : dict
            Custom resources to reserve for the remote function.

        Returns
        -------
        RemoteFunction
        """
        remote_function_cls = self._backend_cls.make_remote_function(
            function, num_cpus, num_returns, resources
        )
        return RemoteFunction(remote_function_cls, function)

    def make_actor(self, cls, num_cpus, resources):
        """
        Define an actor class.

        cls : object
            Class to be an actor class.
        num_cpus : int
            The number of CPUs to reserve for the lifetime of the actor.
        resources : dict
            Custom resources to reserve for the lifetime of the actor.

        Returns
        -------
        ActorClass
        """
        actor_cls, actor_cls_args = self._backend_cls.make_actor(
            cls, num_cpus, resources
        )
        return ActorClass(actor_cls, *actor_cls_args)

    def get(self, object_refs):
        """
        Get a remote object or a list of remote objects
        from distributed memory.

        Parameters
        ----------
        object_refs : ObjectRef or list
            ``ObjectRef`` or a list of ``ObjectRef``-s to get data from.

        Returns
        -------
        object
            A Python object or a list of Python objects.
        """
        if isinstance(object_refs, ObjectRef):
            object_refs = object_refs._ref
        else:
            object_refs = [
                obj_ref._ref if isinstance(obj_ref, ObjectRef) else obj_ref
                for obj_ref in object_refs
            ]
        return self._backend_cls.get(object_refs)

    def put(self, data):
        """
        Put data into distributed memory.

        Parameters
        ----------
        data : object
            Data to be put.

        Returns
        -------
        ObjectRef
            ObjectRef matching to data.
        """
        ref = self._backend_cls.put(data)
        return ObjectRef(ref)

    def wait(self, object_refs, num_returns=1):
        """
        Wait until `object_refs` are finished.

        This method returns two lists. The first list consists of
        ``ObjectRef``-s that correspond to objects that completed computations.
        The second list corresponds to the rest of the ``ObjectRef``-s (which may or may not be ready).

        Parameters
        ----------
        object_refs : ObjectRef or list
            ``ObjectRef`` or list of ``ObjectRef``-s to be waited.
        num_returns : int, default: 1
            The number of ``ObjectRef``-s that should be returned as ready.

        Returns
        -------
        two lists
            List of ``ObjectRef``-s that are ready and list of the remaining ``ObjectRef``-s.
        """
        if isinstance(object_refs, ObjectRef):
            object_refs = [object_refs._ref]
        else:
            object_refs = [
                obj_ref._ref if isinstance(obj_ref, ObjectRef) else obj_ref
                for obj_ref in object_refs
            ]
        ready, not_ready = self._backend_cls.wait(object_refs, num_returns=num_returns)
        ready = [ObjectRef(r) for r in ready]
        not_ready = [ObjectRef(nr) for nr in not_ready]
        return ready, not_ready

    @staticmethod
    def is_object_ref(obj):
        """
        Whether an object is ``ObjectRef`` or not.

        Parameters
        ----------
        obj : object
            An object to be checked.

        Returns
        -------
        bool
            `True` if an object is ``ObjectRef``, `False` otherwise.
        """
        return isinstance(obj, ObjectRef)

    def get_ip(self):
        """
        Get node IP address.

        Returns
        -------
        str
            Node IP address.
        """
        return self._backend_cls.get_ip()

    def num_cpus(self):
        """
        Get the number of CPUs used by the execution backend.

        Returns
        -------
        int
        """
        return self._backend_cls.num_cpus()

    def cluster_resources(self):
        """
        Get resources of the cluster.

        Returns
        -------
        dict
            Dictionary with cluster nodes info in the form
            `{"node_ip0": {"CPU": x0}, "node_ip1": {"CPU": x1}, ...}`.
        """
        return self._backend_cls.cluster_resources()

    def shutdown(self):
        """Shutdown an execution backend."""
        self._backend_cls.shutdown()

    def is_initialized(self):
        """
        Check if a unidist backend has already been initialized.

        Returns
        -------
        bool
            True or False.
        """
        return self._backend_cls.is_initialized()



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Core base remote function specific functionality."""

from functools import wraps

from .common import filter_arguments


class RemoteFunction:
    """
    A class that is base for any remote function class specific for the backend.

    This wraps an instance of the child class, which in turn wraps
    a remote function that is meant to be remote, specific for the backend.

    Parameters
    ----------
    remote_function_cls : RemoteFunction
        An instance of the child class.
    function : callable
        A function that is meant to be remote.
    """

    def __init__(self, remote_function_cls, function):
        self._remote_function_cls = remote_function_cls
        self._original_function = function

        # Override task.remote's signature and docstring
        @wraps(self._original_function)
        def _remote_proxy(*args, **kwargs):
            args, kwargs = filter_arguments(*args, **kwargs)

            return self._remote_function_cls._remote(*args, **kwargs)

        self.remote = _remote_proxy

    def __call__(self, *args, **kwargs):
        raise TypeError(
            "Remote functions cannot be called directly. Instead "
            f"of running '{self._original_function.__name__}()', "
            f"try '{self._original_function.__name__}.remote()'."
        )

    def remote(self, *args, **kwargs):
        """
        Call a remote function in a worker process.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be passed in the remote function.
        **kwargs : dict
            Keyword arguments to be passed in the remote function.

        Returns
        -------
        ObjectRef, list or None
        """
        return self._remote_function_cls._remote(*args, **kwargs)

    def options(self, *args, num_cpus=None, num_returns=None, resources=None, **kwargs):
        """
        Override the remote function invocation parameters.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be overrided.
        num_cpus : int, optional
            Positional arguments to be overrided.
        num_returns : int, optional
            Positional arguments to be overrided.
        resources : int, optional
            Positional arguments to be overrided.
        **kwargs : dict
            Keyword arguments to be overrided.

        Returns
        -------
        FuncWrapper
            An instance of wrapped function that a non-underscore .remote() can be called on.
        """
        remote_function_cls = self._remote_function_cls

        class FuncWrapper:
            def remote(self, *args, **kwargs):
                args, kwargs = filter_arguments(*args, **kwargs)

                return remote_function_cls._remote(
                    *args,
                    num_cpus=num_cpus,
                    num_returns=num_returns,
                    resources=resources,
                    **kwargs,
                )

        return FuncWrapper()



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Actor specific functionality implemented using Python Multiprocessing."""

import cloudpickle as pkl
from multiprocessing.managers import BaseManager

from unidist.core.backends.pymp.core.object_store import ObjectStore, Delayed
from unidist.core.backends.pymp.core.process_manager import (
    ProcessManager,
    Task,
)


class ActorMethod:
    """
    Class is responsible to execute `method_name` of
    `cls_obj` in the separate worker-process of `actor` object.

    Parameters
    ----------
    cls_obj : multiprocessing.managers.BaseManager
        Shared manager-class.
    actor : Actor
        Actor object.
    method_name : str
        The name of the method to be called.
    obj_store : unidist.core.backends.pymp.core.object_store.ObjectStore
        Object storage to share data between workers.
    """

    def __init__(self, cls_obj, actor, method_name, obj_store):
        self._cls_obj = cls_obj
        self._method_name = method_name
        self._actor = actor
        self._obj_store = obj_store

    def submit(self, *args, num_returns=1, **kwargs):
        """
        Execute `self._method_name` asynchronously in the worker of `self._actor`.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be passed in the `self._method_name` method.
        num_returns : int, default: 1
            Number of results to be returned from `self._method_name`.
        **kwargs : dict
            Keyword arguments to be passed in the `self._method_name` method.

        Returns
        -------
        unidist.core.backends.common.data_id.DataID, list or None
            Type of returns depends on `num_returns` value:

            * if `num_returns == 1`, ``DataID`` will be returned.
            * if `num_returns > 1`, list of ``DataID``-s will be returned.
            * if `num_returns == 0`, ``None`` will be returned.
        """
        if num_returns == 0:
            data_ids = None
        elif num_returns > 1:
            data_ids = [self._obj_store.put(Delayed()) for _ in range(num_returns)]
        else:
            data_ids = self._obj_store.put(Delayed())

        cls_method = getattr(self._cls_obj, self._method_name)

        task = Task(cls_method, data_ids, self._obj_store, *args, **kwargs)
        self._actor.submit(task)

        return data_ids


class Actor:
    """
    Actor-class to execute methods of wrapped class in a separate worker.

    Parameters
    ----------
    cls : object
        Class to be an actor class.
    *args : iterable
        Positional arguments to be passed in `cls` constructor.
    **kwargs : dict
        Keyword arguments to be passed in `cls` constructor.

    Notes
    -----
    Python multiprocessing manager-class will be created to wrap `cls`.
    This makes `cls` class object shared between different workers. Manager-class
    starts additional process to share class state between processes.

    Methods of `cls` class object are executed in the worker, grabbed from a workers pool.
    """

    def __init__(self, cls, *args, **kwargs):
        self._worker = None
        self._worker_id = None
        self._obj_store = ObjectStore.get_instance()

        # FIXME : Change "WrappedClass" -> cls.__name__ + "Manager", for example.
        BaseManager.register("WrappedClass", cls)
        manager = BaseManager()
        manager.start()

        self._cls_obj = manager.WrappedClass(*args, **kwargs)
        self._worker, self._worker_id = ProcessManager.get_instance().grab_worker()

    def __getattr__(self, name):
        """
        Get the attribute `name` of the `self._cls_obj` class.

        This methods creates the ``ActorMethod`` object that is responsible
        for calling method `name` of the `self._cls_obj` class remotely.

        Parameters
        ----------
        name : str
            Name of the method to be called remotely.

        Returns
        -------
        ActorMethod
        """
        return ActorMethod(self._cls_obj, self, name, self._obj_store)

    def submit(self, task):
        """
        Execute `task` asynchronously in the worker grabbed by this actor.

        Parameters
        ----------
        task : unidist.core.backends.pymp.core.process_manager.Task
            Task object holding callable function.
        """
        self._worker.add_task(pkl.dumps(task))

    def __del__(self):
        """
        Destructor of the actor.

        Free worker, grabbed from the workers pool.
        """
        if self._worker_id is not None:
            ProcessManager.get_instance().free_worker(self._worker_id)



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

from collections import defaultdict

import unidist.core.backends.mpi.core.common as common
import unidist.core.backends.mpi.core.communication as communication
from unidist.core.backends.mpi.core.controller.common import push_data
from unidist.core.backends.mpi.core.object_store import ObjectStore


mpi_state = communication.MPIState.get_instance()
# Logger configuration
# When building documentation we do not have MPI initialized so
# we use the condition to set "worker_0.log" in order to build it succesfully.
logger_name = "worker_{}".format(mpi_state.global_rank if mpi_state is not None else 0)
log_file = "{}.log".format(logger_name)
logger = common.get_logger(logger_name, log_file)


class RequestStore:
    """
    Class that stores data requests that couldn't be satisfied now.

    Attributes
    ----------
    GET : int, default 0
        `Get` request from other worker to be executed.
    WAIT : int, default 1
        `Wait` request from other worker to be executed.
    DATA : int, default 2
        Data request to other worker.

    Notes
    -----
    Supports `GET` and `WAIT` requests.
    """

    __instance = None

    GET = 0
    WAIT = 1
    DATA = 2

    def __init__(self):
        # Non-blocking get requests {DataID : [ Set of Ranks ]}
        self._nonblocking_get_requests = defaultdict(set)
        # Blocking get requests {DataID : [ Set of Ranks ]}
        self._blocking_get_requests = defaultdict(set)
        # Blocking wait requests {DataID : Rank}
        self._blocking_wait_requests = {}
        # Data requests
        self._data_requests = set()

    @classmethod
    def get_instance(cls):
        """
        Get instance of ``RequestStore``.

        Returns
        -------
        RequestStore
        """
        if cls.__instance is None:
            cls.__instance = RequestStore()
        return cls.__instance

    def put(self, data_id, rank, request_type, is_blocking_op=False):
        """
        Save request type for this data ID for later processing.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.
        rank : int
            Source rank requester.
        request_type : int
            Type of request.
        is_blocking_op : bool
            Whether the get request should be blocking or not.
            If ``True``, the request should be processed immediatly
            even for a worker since it can get into controller mode.
        """
        if request_type == self.GET:
            if is_blocking_op:
                self._blocking_get_requests[data_id].add(rank)
            else:
                self._nonblocking_get_requests[data_id].add(rank)
        elif request_type == self.WAIT:
            self._blocking_wait_requests[data_id] = rank
        elif request_type == self.DATA:
            self._data_requests.add(data_id)
        else:
            raise ValueError("Unsupported request type option!")

    def is_data_already_requested(self, data_id):
        """
        Check if data by particular `data_id` was already requested from another MPI process.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.

        Returns
        -------
        bool
            ``True`` if communnication request was happened for this ID.
        """
        return data_id in self._data_requests

    def discard_data_request(self, data_id):
        """
        Discard data request by `data_id` because the data has become available.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.
        """
        self._data_requests.discard(data_id)

    def clear_get_requests(self):
        """Clear blocking and non-blocking get requests."""
        self._blocking_get_requests.clear()
        self._nonblocking_get_requests.clear()

    def clear_wait_requests(self):
        """Clear blocking wait requests requests."""
        self._blocking_wait_requests.clear()

    def check_pending_get_requests(self, data_ids):
        """
        Check if `GET` event on this `data_ids` is waiting to be processed.

        Process the request if data ID available in local object store.

        Parameters
        ----------
        data_id : iterable or unidist.core.backends.mpi.core.common.MpiDataID
            An ID or list of IDs to data.
        """

        def check_request(data_id):
            # Check non-blocking data requests for one of the workers
            if data_id in self._nonblocking_get_requests:
                ranks_with_get_request = self._nonblocking_get_requests[data_id]
                for rank_num in ranks_with_get_request:
                    # Data is already in DataMap, so not problem here
                    self.process_get_request(rank_num, data_id, is_blocking_op=False)
                del self._nonblocking_get_requests[data_id]
            # Check blocking data requests for other of the workers
            if data_id in self._blocking_get_requests:
                ranks_with_get_request = self._blocking_get_requests[data_id]
                for rank_num in ranks_with_get_request:
                    # Data is already in DataMap, so not problem here
                    self.process_get_request(rank_num, data_id, is_blocking_op=True)
                del self._blocking_get_requests[data_id]

        if isinstance(data_ids, (list, tuple)):
            for data_id in data_ids:
                check_request(data_id)
        else:
            check_request(data_ids)

    def check_pending_wait_requests(self, data_ids):
        """
        Check if `WAIT` event on this `data_ids` is waiting to be processed.

        Process the request if data ID available in local object store.
        Send signal without any data.

        Parameters
        ----------
        data_id : iterable or unidist.core.backends.mpi.core.common.MpiDataID
            An ID or list of IDs to data.
        """
        if isinstance(data_ids, (list, tuple)):
            for data_id in data_ids:
                if data_id in self._blocking_wait_requests:
                    # Data is already in DataMap, so not problem here.
                    # We use a blocking send here because the receiver is waiting for the result.
                    communication.mpi_send_object(
                        communication.MPIState.get_instance().global_comm,
                        data_id,
                        communication.MPIRank.ROOT,
                    )
                    del self._blocking_wait_requests[data_id]
        else:
            if data_ids in self._blocking_wait_requests:
                # We use a blocking send here because the receiver is waiting for the result.
                communication.mpi_send_object(
                    communication.MPIState.get_instance().global_comm,
                    data_ids,
                    communication.MPIRank.ROOT,
                )
                del self._blocking_wait_requests[data_ids]

    def process_wait_request(self, data_id):
        """
        Satisfy WAIT operation request from another process.

        Save request for later processing if `data_id` is not available currently.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            Chech if `data_id` is available in object store.

        Notes
        -----
        Only ROOT rank is supported for now, therefore no rank argument needed.
        """
        object_store = ObjectStore.get_instance()
        if object_store.contains(data_id):
            # Executor wait just for signal
            # We use a blocking send here because the receiver is waiting for the result.
            communication.mpi_send_object(
                communication.MPIState.get_instance().global_comm,
                data_id,
                communication.MPIRank.ROOT,
            )
            logger.debug("Wait data {} id is ready".format(data_id._id))
        else:
            self.put(data_id, communication.MPIRank.ROOT, self.WAIT)
            logger.debug("Pending wait request {} id".format(data_id._id))

    def process_get_request(self, source_rank, data_id, is_blocking_op=False):
        """
        Satisfy GET operation request from another process.

        Save request for later processing if `data_id` is not available currently.

        Parameters
        ----------
        source_rank : int
            Rank number to send data to.
        data_id: unidist.core.backends.mpi.core.common.MpiDataID
            `data_id` associated data to request
        is_blocking_op : bool, default: False
            Whether the get request should be blocking or not.
            If ``True``, the request should be processed immediatly
            even for a worker since it can get into controller mode.

        Notes
        -----
        Request is asynchronous, no wait for the data sending.
        """
        object_store = ObjectStore.get_instance()
        if object_store.contains(data_id):
            push_data(
                source_rank,
                data_id,
                is_blocking_op=source_rank == communication.MPIRank.ROOT
                or is_blocking_op,
            )
            logger.debug(
                "Send requested {} id to {} rank - PROCESSED".format(
                    data_id._id, source_rank
                )
            )
        else:
            logger.debug(
                "Pending request {} id to {} rank".format(data_id._id, source_rank)
            )
            self.put(data_id, source_rank, self.GET, is_blocking_op=is_blocking_op)



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Utilities used to initialize MPI execution backend."""


def initialize_mpi():
    """Initialize the MPI execution backend."""
    from unidist.core.backends.mpi.core import init

    init()


class ImmutableDict(dict):
    __readonly_exception = TypeError("Cannot modify `ImmutableDict`")

    def __setitem__(self, *args, **kwargs):
        raise ImmutableDict.__readonly_exception

    def __delitem__(self, *args, **kwargs):
        raise ImmutableDict.__readonly_exception

    def pop(self, *args, **kwargs):
        raise ImmutableDict.__readonly_exception

    def popitem(self, *args, **kwargs):
        raise ImmutableDict.__readonly_exception

    def clear(self, *args, **kwargs):
        raise ImmutableDict.__readonly_exception

    def update(self, *args, **kwargs):
        raise ImmutableDict.__readonly_exception

    def setdefault(self, *args, **kwargs):
        raise ImmutableDict.__readonly_exception



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Serialization interface"""

import importlib
import inspect
import sys
from collections.abc import KeysView

try:
    import mpi4py
except ImportError:
    raise ImportError(
        "Missing dependency 'mpi4py'. Use pip or conda to install it."
    ) from None

# Serialization libraries
if sys.version_info[1] < 8:  # check the minor Python version
    try:
        import pickle5 as pkl
    except ImportError:
        raise ImportError(
            "Missing dependency 'pickle5'. Use pip or conda to install it."
        ) from None
else:
    import pickle as pkl
import cloudpickle as cpkl
import msgpack
import gc  # msgpack optimization

from unidist.config import MpiPickleThreshold

# TODO: Find a way to move this after all imports
mpi4py.rc(recv_mprobe=False, initialize=False)
from mpi4py.MPI import memory  # noqa: E402


# Pickle 5 protocol compatible types check
compatible_modules = ("pandas", "numpy")
available_modules = []
for module_name in compatible_modules:
    try:
        available_modules.append(importlib.import_module(module_name))
    except ModuleNotFoundError:
        pass


def is_cpkl_serializable(data):
    """
    Check if the data should be serialized with cloudpickle.

    Parameters
    ----------
    data : object
        Python object

    Returns
    -------
    bool
        ``True` if the data should be serialized with cloudpickle library.
    """
    return (
        inspect.isfunction(data)
        or inspect.isclass(data)
        or inspect.ismethod(data)
        or data.__class__.__module__ != "builtins"
        or isinstance(data, KeysView)
    )


def is_pickle5_serializable(data):
    """
    Check if the data should be serialized with pickle 5 protocol.

    Parameters
    ----------
    data : object
        Python object.

    Returns
    -------
    bool
        ``True`` if the data should be serialized with pickle using protocol 5 (out-of-band data).
    """
    for module in available_modules:
        if module.__name__ == "pandas" and isinstance(
            data, (module.DataFrame, module.Series)
        ):
            return True
        elif module.__name__ == "numpy" and isinstance(data, module.ndarray):
            return True
    return False


class ComplexDataSerializer:
    """
    Class for data serialization/de-serialization for MPI comminication.

    Parameters
    ----------
    buffers : list, default: None
        A list of ``PickleBuffer`` objects for data decoding.
    buffer_count : list, default: None
        List of the number of buffers for each object
        to be serialized/deserialized using the pickle 5 protocol.

    Notes
    -----
    Uses a combination of msgpack, cloudpickle and pickle libraries.
    Msgpack allows to serialize/deserialize internal objects of a container separately,
    but send them as one object. For example, for an array of pandas DataFrames,
    each DataFrame will be serialized separately using pickle 5,
    and all `buffers` will be stored in one array to be sent together.
    To deserialize it `buffer_count` is used, which contains information
    about the number of `buffers` for each internal object.
    """

    # Minimum buffer size for serialization with pickle 5 protocol
    PICKLE_THRESHOLD = MpiPickleThreshold.get()

    def __init__(self, buffers=None, buffer_count=None):
        self.buffers = buffers if buffers else []
        self.buffer_count = list(buffer_count) if buffer_count else []
        self._callback_counter = 0

    def _buffer_callback(self, pickle_buffer):
        """
        Callback for pickle protocol 5 out-of-band data buffers collection.

        Parameters
        ----------
        pickle_buffer: pickle.PickleBuffer
            Pickle library buffer wrapper.
        """
        pickle_buffer = memory(pickle_buffer)
        if len(pickle_buffer) >= self.PICKLE_THRESHOLD:
            self.buffers.append(pickle_buffer)
            self._callback_counter += 1
            return False
        return True

    def _dataframe_encode(self, frame):
        """
        Encode with pickle library using protocol 5.

        Parameters
        ----------
        data : object
            Pickle 5 serializable object (e.g. pandas DataFrame or NumPy array).

        Returns
        -------
        dict
            Dictionary with array of serialized bytes.
        """
        s_frame = pkl.dumps(frame, protocol=5, buffer_callback=self._buffer_callback)
        self.buffer_count.append(self._callback_counter)
        self._callback_counter = 0
        return {"__pickle5_custom__": True, "as_bytes": s_frame}

    def _cpkl_encode(self, obj):
        """
        Encode with cloudpickle library.

        Parameters
        ----------
        obj : object
            Python object.

        Returns
        -------
        dict
            Dictionary with array of serialized bytes.
        """
        return {"__cloud_custom__": True, "as_bytes": cpkl.dumps(obj)}

    def _pkl_encode(self, obj):
        """
        Encode with pickle library.

        Parameters
        ----------
        obj : object
            Python object.

        Returns
        -------
        dict
            Dictionary with array of serialized bytes.
        """
        return {"__pickle_custom__": True, "as_bytes": pkl.dumps(obj)}

    def _encode_custom(self, obj):
        """
        Serialization hook for msgpack library.

        It encodes complex data types the library couldn`t handle.

        Parameters
        ----------
        obj : object
            Python object.
        """
        if is_pickle5_serializable(obj):
            return self._dataframe_encode(obj)
        elif is_cpkl_serializable(obj):
            return self._cpkl_encode(obj)
        else:
            return self._pkl_encode(obj)

    def serialize(self, data):
        """
        Serialize data to a byte array.

        Parameters
        ----------
        data : object
            Data to serialize.

        Returns
        -------
        bytes
            Serialized data.

        Notes
        -----
        Uses msgpack, cloudpickle and pickle libraries.
        """
        return msgpack.packb(data, default=self._encode_custom, strict_types=True)

    def _decode_custom(self, obj):
        """
        De-serialization hook for msgpack library.

        It decodes complex data types the library couldn`t handle.

        Parameters
        ----------
        obj : object
            Python object.
        """
        if "__cloud_custom__" in obj:
            return cpkl.loads(obj["as_bytes"])
        elif "__pickle_custom__" in obj:
            return pkl.loads(obj["as_bytes"])
        elif "__pickle5_custom__" in obj:
            frame = pkl.loads(obj["as_bytes"], buffers=self.buffers)
            # check if there are out-of-band buffers
            # TODO: look at this condition and get rid of it because
            # `buffer_count` should always be a list with length greater 0.
            if self.buffer_count:
                del self.buffers[: self.buffer_count.pop(0)]
            return frame
        else:
            return obj

    def deserialize(self, s_data):
        """
        De-serialize data from a bytearray.

        Parameters
        ----------
        s_data : bytearray
            Data to de-serialize.

        Returns
        -------
        object
            Deserialized data.

        Notes
        -----
        Uses msgpack, cloudpickle and pickle libraries.
        """
        gc.disable()  # Performance optimization for msgpack
        unpacked_data = msgpack.unpackb(
            s_data, object_hook=self._decode_custom, strict_map_key=False
        )
        gc.enable()
        return unpacked_data


class SimpleDataSerializer:
    """
    Class for simple data serialization/de-serialization for MPI communication.

    Notes
    -----
    Uses cloudpickle and pickle libraries as separate APIs.
    """

    def serialize_cloudpickle(self, data):
        """
        Encode with a cloudpickle library.

        Parameters
        ----------
        obj : object
            Python object.

        Returns
        -------
        bytearray
            Array of serialized bytes.
        """
        return cpkl.dumps(data)

    def serialize_pickle(self, data):
        """
        Encode with a pickle library.

        Parameters
        ----------
        obj : object
            Python object.

        Returns
        -------
        bytearray
            Array of serialized bytes.
        """
        return pkl.dumps(data)

    def deserialize_cloudpickle(self, data):
        """
        De-serialization with cloudpickle library.

        Parameters
        ----------
        obj : bytearray
            Python object.

        Returns
        -------
        object
            Original reconstructed object.
        """
        return cpkl.loads(data)

    def deserialize_pickle(self, data):
        """
        De-serialization with pickle library.

        Parameters
        ----------
        obj : bytearray
            Python object.

        Returns
        -------
        object
            Original reconstructed object.
        """
        return pkl.loads(data)


def serialize_complex_data(data):
    """
    Serialize data to a bytearray.

    Parameters
    ----------
    data : object
        Data to serialize.

    Returns
    -------
    bytes
        Serialized data.

    Notes
    -----
    Uses msgpack, cloudpickle and pickle libraries.
    """
    serializer = ComplexDataSerializer()
    s_data = serializer.serialize(data)
    serialized_data = {
        "s_data": s_data,
        "raw_buffers": serializer.buffers,
        "buffer_count": serializer.buffer_count,
    }
    return serialized_data


def deserialize_complex_data(s_data, raw_buffers, buffer_count):
    """
    Deserialize data based on passed in information.

    Parameters
    ----------
    s_data : bytearray
        Serialized msgpack data.
    raw_buffers : list
        A list of ``PickleBuffer`` objects for data decoding.
    buffer_count : list
        List of the number of buffers for each object
        to be deserialized using the pickle 5 protocol.

    Returns
    -------
    object
        Deserialized data.
    Notes
    -----
    Uses msgpack, cloudpickle and pickle libraries.
    """
    deserializer = ComplexDataSerializer(raw_buffers, buffer_count)
    return deserializer.deserialize(s_data)



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Actor specific functionality using Dask backend."""

from distributed.client import get_client

from unidist.core.base.actor import Actor, ActorMethod
from unidist.core.base.object_ref import ObjectRef
from unidist.core.backends.common.utils import unwrap_object_refs


class DaskActorMethod(ActorMethod):
    """
    The class implements the interface in ``ActorMethod`` using Dask backend.

    Parameters
    ----------
    cls : distributed.actor.Actor
        An actor class from which method `method_name` will be called.
    method_name : str
        The name of the method to be called.
    """

    def __init__(self, cls, method_name):
        self._cls = cls
        self._method_name = method_name
        self._num_returns = 1

    def _remote(self, *args, num_returns=None, **kwargs):
        """
        Execute `self._method_name` in a worker process.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be passed in the method.
        num_returns : int, optional
            Number of results to be returned. If it isn't
            provided, `self._num_returns` will be used.
        **kwargs : dict
            Keyword arguments to be passed in the method.

        Returns
        -------
        ObjectRef, list or None
            Type of returns depends on `num_returns` value:

            * if `num_returns == 1`, ``ObjectRef`` will be returned.
            * if `num_returns > 1`, list of ``ObjectRef`` will be returned.
            * if `num_returns == 0`, ``None`` will be returned.
        """
        client = get_client()

        if num_returns is None:
            num_returns = self._num_returns

        # FIXME: We are currently unwrapping all the nested Dask Future objects from `args` and `kwargs`
        # to pass those in `self._method_name` because of Dask Distributed issue #4842

        unwrapped_args = tuple()
        for arg in args:
            unwrapped_args += (unwrap_object_refs(arg),)

        unwrapped_kwargs = dict()
        for k, v in kwargs.items():
            unwrapped_kwargs[k] = unwrap_object_refs(v)

        future = client.submit(
            lambda actor_cls, method, *args, **kwargs: getattr(actor_cls, method)(
                *args, **kwargs
            ),
            self._cls,
            self._method_name,
            *unwrapped_args,
            **unwrapped_kwargs,
            pure=False,
        )

        if num_returns == 1:
            return ObjectRef(client.submit(lambda f: f.result(), future, pure=False))
        elif num_returns > 1:
            return [
                ObjectRef(
                    client.submit(lambda f, i: f.result()[i], future, i, pure=False)
                )
                for i in range(num_returns)
            ]
        elif num_returns == 0:
            return None


class DaskActor(Actor):
    """
    The class implements the interface in ``Actor`` using Dask backend.

    Parameters
    ----------
    cls : object
        Class to be an actor class.
    num_cpus : int
        The number of CPUs to reserve for the lifetime of the actor.
    resources : dict
        Custom resources to reserve for the lifetime of the actor.
    actor_handle : None or distributed.actor.Actor
        Used for proper serialization/deserialization via `__reduce__`.
    """

    def __init__(self, cls, num_cpus, resources, actor_handle=None):
        self._cls = cls
        self._num_cpus = num_cpus
        self._resources = resources
        self._actor_handle = actor_handle

    def _serialization_helper(self):
        """
        Helper to save the state of the object.

        This is defined to make pickling work via `__reduce__`.

        Returns
        -------
        dict
            A dictionary of the information needed to reconstruct the object.
        """
        state = {
            "cls": self._cls,
            "num_cpus": self._num_cpus,
            "resources": self._resources,
            "actor_handle": self._actor_handle,
        }
        return state

    @classmethod
    def _deserialization_helper(cls, state):
        """
        Helper to restore the state of the object.

        This is defined to make pickling work via `__reduce__`.

        Parameters
        ----------
        state : dict
            The serialized state of the object.

        Returns
        -------
        DaskActor
        """
        return cls(
            state["cls"],
            state["num_cpus"],
            state["resources"],
            actor_handle=state["actor_handle"],
        )

    def __reduce__(self):
        """
        This is defined intentionally to make pickling work correctly.

        Returns
        -------
        tuple
            Callable and arguments to be passed in to it.
        """
        state = self._serialization_helper()
        return self._deserialization_helper, (state,)

    def __getattr__(self, name):
        """
        Get the attribute `name` of the `self._cls` class.

        This methods creates the ``DaskActorMethod`` object that is responsible
        for calling method `name` of the `self._cls` class remotely.

        Parameters
        ----------
        name : str
            Name of the method to be called remotely.

        Returns
        -------
        DaskActorMethod
        """
        return DaskActorMethod(self._actor_handle, name)

    def _remote(self, *args, num_cpus=None, resources=None, **kwargs):
        """
        Create actor class, specific for Dask backend, from `self._cls`.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be passed in `self._cls` class constructor.
        num_cpus : int, optional
            The number of CPUs to reserve for the lifetime of the actor.
        resources : dict, optional
            Custom resources to reserve for the lifetime of the actor.
        **kwargs : dict
            Keyword arguments to be passed in `self._cls` class constructor.

        Returns
        -------
        DaskActor
        """
        if num_cpus is not None or self._num_cpus is not None:
            raise NotImplementedError(
                "'num_cpus' is not supported yet by Dask backend."
            )
        if resources is not None or self._resources is not None:
            raise NotImplementedError(
                "'resources' is not supported yet by Dask backend."
            )

        client = get_client()

        self._actor_handle = client.submit(
            self._cls, *args, **kwargs, actor=True
        ).result()
        return self



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Commons used by core base functionality."""

from .object_ref import ObjectRef


class BackendName:
    """String representations of unidist backends."""

    RAY = "ray"
    MPI = "mpi"
    DASK = "dask"
    PYMP = "pymp"
    PYSEQ = "pyseq"


def filter_arguments(*args, **kwargs):
    """
    Filter `args` and `kwargs` so that a backend itself is able to materialize data.

    Parameters
    ----------
    *args : list or tuple
        Positional arguments to be filtered.
    **kwargs : dict
        Keyword arguments to be filtered.

    Returns
    -------
    list
        Filtered positional arguments.
    dict
        Filtered keyword arguments.

    Notes
    -----
    The method unwraps one level of nesting for `args` and `kwargs` so that
    an underlying backend is able to materialize data.
    Data materialization of the next levels of nesting is a user's burden.
    """
    args = [] if args is None else args
    kwargs = {} if kwargs is None else kwargs

    args = [arg._ref if isinstance(arg, ObjectRef) else arg for arg in args]
    kwargs = {k: (v._ref if isinstance(v, ObjectRef) else v) for k, v in kwargs.items()}
    return args, kwargs



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

from libc.stdint cimport uint8_t, int64_t

cimport memory

def parallel_memcopy(const uint8_t[:] src, uint8_t[:] dst, int memcopy_threads):
    """
    Multithreaded data copying between buffers.

    Parameters
    ----------
    src : uint8_t[:]
        Copied data.
    dst : uint8_t[:]
        Buffer for writing.
    memcopy_threads : int
        Number of threads to write.
    """
    with nogil:
        memory.parallel_memcopy(&dst[0],
                                &src[0],
                                len(src),
                                64,
                                memcopy_threads)

def fill(int64_t[:] buff, int64_t value):
    """
    Fill a given buffer with a given value.

    Parameters
    ----------
    buff : int64_t[:]
        Original data.
    value : int64_t
        Value to fill.
    """
    with nogil:
        memory.fill(&buff[0], len(buff), value)



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Worker MPI process task processing functionality."""

import asyncio
from functools import wraps, partial

try:
    import mpi4py
except ImportError:
    raise ImportError(
        "Missing dependency 'mpi4py'. Use pip or conda to install it."
    ) from None

import unidist.core.backends.mpi.core.common as common
import unidist.core.backends.mpi.core.communication as communication
from unidist.core.backends.mpi.core.object_store import ObjectStore
from unidist.core.backends.mpi.core.local_object_store import LocalObjectStore
from unidist.core.backends.mpi.core.worker.request_store import RequestStore
from unidist.core.backends.mpi.core.worker.task_store import TaskStore
from unidist.core.backends.mpi.core.async_operations import AsyncOperations
from unidist.core.backends.mpi.core.controller.common import pull_data
from unidist.core.backends.mpi.core.shared_object_store import SharedObjectStore

# TODO: Find a way to move this after all imports
mpi4py.rc(recv_mprobe=False, initialize=False)
from mpi4py import MPI  # noqa: E402

mpi_state = communication.MPIState.get_instance()
# Logger configuration
# When building documentation we do not have MPI initialized so
# we use the condition to set "worker_0.log" in order to build it succesfully.
logger_name = "worker_{}".format(mpi_state.global_rank if mpi_state is not None else 0)
log_file = "{}.log".format(logger_name)
w_logger = common.get_logger(logger_name, log_file)
# Actors map {handle : actor}
actor_map = {}


# -------------- #
# Loop utilities #
# -------------- #


def async_wrap(func):
    """
    Make a decorator that allows to run a regular function asynchronously.

    Parameters
    ----------
    func : callable
        The function, which should be called like as coroutine.

    Returns
    -------
    callable
        Function-decorator.
    """

    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        pfunc = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, pfunc)

    return run


# ---------------------------- #
# Worker event processing loop #
# ---------------------------- #


async def worker_loop():
    """
    Infinite operations processing loop.

    Master or any worker could be the source of an operation.

    Notes
    -----
    The loop exits on special cancelation operation.
    ``unidist.core.backends.mpi.core.common.Operations`` defines a set of supported operations.
    """
    task_store = TaskStore.get_instance()
    object_store = ObjectStore.get_instance()
    local_store = LocalObjectStore.get_instance()
    request_store = RequestStore.get_instance()
    async_operations = AsyncOperations.get_instance()

    # Barrier to check if worker process is ready to start the communication loop
    mpi_state.global_comm.Barrier()
    w_logger.debug("Worker loop started")
    # Once the worker receives the cancel signal from ``Root`` rank,
    # it is getting to shutdown. All pending requests and communications are cancelled,
    # and the worker sends the ready to shutdown signal to ``Monitor``.
    # Once all workers excluding ``Root`` and ``Monitor`` ranks are ready to shutdown,
    # ``Monitor` sends the shutdown signal to every worker so they can exit the loop.
    ready_to_shutdown_posted = False
    while True:
        # Listen receive operation from any source
        operation_type, source_rank = await async_wrap(
            communication.mpi_recv_operation
        )(mpi_state.global_comm)
        w_logger.debug(
            f"common.Operation processing - {operation_type} from {source_rank} rank"
        )

        # Proceed the request
        if operation_type == common.Operation.EXECUTE:
            request = pull_data(mpi_state.global_comm, source_rank)
            if not ready_to_shutdown_posted:
                # Execute the task if possible
                pending_request = task_store.process_task_request(request)
                if pending_request:
                    task_store.put(pending_request)
                else:
                    # Check pending requests. Maybe some data became available.
                    task_store.check_pending_tasks()

        elif operation_type == common.Operation.GET:
            request = communication.mpi_recv_object(mpi_state.global_comm, source_rank)
            if request is not None and not ready_to_shutdown_posted:
                request_store.process_get_request(
                    request["source"], request["id"], request["is_blocking_op"]
                )

        elif operation_type == common.Operation.PUT_DATA:
            request = pull_data(mpi_state.global_comm, source_rank)
            if not ready_to_shutdown_posted:
                w_logger.debug(
                    "PUT (RECV) {} id from {} rank".format(
                        request["id"]._id, source_rank
                    )
                )

                # Discard data request to another worker, if data has become available
                request_store.discard_data_request(request["id"])

                # Check pending requests. Maybe some data became available.
                task_store.check_pending_tasks()
                # Check pending actor requests also.
                task_store.check_pending_actor_tasks()

        elif operation_type == common.Operation.PUT_OWNER:
            request = communication.mpi_recv_object(mpi_state.global_comm, source_rank)
            if not ready_to_shutdown_posted:
                local_store.put_data_owner(request["id"], request["owner"])

                w_logger.debug(
                    "PUT_OWNER {} id is owned by {} rank".format(
                        request["id"]._id, request["owner"]
                    )
                )

        elif operation_type == common.Operation.PUT_SHARED_DATA:
            result = pull_data(mpi_state.global_comm, source_rank)

            # Clear cached request to another worker, if data_id became available
            request_store.discard_data_request(result["id"])

            # Check pending requests. Maybe some data became available.
            task_store.check_pending_tasks()
            # Check pending actor requests also.
            task_store.check_pending_actor_tasks()

        elif operation_type == common.Operation.WAIT:
            request = communication.mpi_recv_object(mpi_state.global_comm, source_rank)
            if not ready_to_shutdown_posted:
                w_logger.debug("WAIT for {} id".format(request["id"]._id))
                request_store.process_wait_request(request["id"])

        elif operation_type == common.Operation.ACTOR_CREATE:
            request = pull_data(mpi_state.global_comm, source_rank)
            if not ready_to_shutdown_posted:
                cls = request["class"]
                args = request["args"]
                kwargs = request["kwargs"]
                handler = request["handler"]
                actor_map[handler] = cls(*args, **kwargs)

        elif operation_type == common.Operation.ACTOR_EXECUTE:
            request = pull_data(mpi_state.global_comm, source_rank)
            if not ready_to_shutdown_posted:
                # Prepare the data
                # Actor method here is a data id so we have to retrieve it from the storage
                method_name = object_store.get(request["task"])
                handler = request["handler"]
                actor_method = getattr(actor_map[handler], method_name)
                request["task"] = actor_method

                # Execute the actor task if possible
                pending_actor_request = task_store.process_task_request(request)
                if pending_actor_request:
                    task_store.put_actor(pending_actor_request)
                else:
                    # Check pending requests. Maybe some data became available.
                    task_store.check_pending_actor_tasks()

        elif operation_type == common.Operation.CLEANUP:
            cleanup_list = communication.recv_serialized_data(
                mpi_state.global_comm, source_rank
            )
            cleanup_list = [common.MpiDataID(*tpl) for tpl in cleanup_list]
            local_store.clear(cleanup_list)

        elif operation_type == common.Operation.CANCEL:
            task_store.clear_pending_tasks()
            task_store.clear_pending_actor_tasks()
            request_store.clear_get_requests()
            request_store.clear_wait_requests()
            async_operations.finish()
            communication.mpi_send_operation(
                mpi_state.global_comm,
                common.Operation.READY_TO_SHUTDOWN,
                mpi_state.get_monitor_by_worker_rank(communication.MPIRank.ROOT),
            )
            ready_to_shutdown_posted = True
        elif operation_type == common.Operation.SHUTDOWN and ready_to_shutdown_posted:
            w_logger.debug("Exit worker event loop")
            SharedObjectStore.get_instance().finalize()
            if not MPI.Is_finalized():
                MPI.Finalize()
            break  # leave event loop and shutdown worker
        else:
            raise ValueError(f"Unsupported operation: {operation_type}")

        # Check completion status of previous async MPI routines
        if not ready_to_shutdown_posted:
            async_operations.check()



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Python Multiprocessing backend core functionality."""

from .actor import Actor
from .api import put, wait, get, submit, init, is_initialized

__all__ = ["Actor", "put", "wait", "get", "submit", "init", "is_initialized"]



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Shared object storage related functionality."""

import cloudpickle as pkl
from multiprocessing import Manager

from unidist.core.backends.common.data_id import DataID


class Delayed:
    """Class-type that used for replacement objects during computation of those."""

    pass


class ObjectStore:
    """
    Class that stores objects and provides access to these from different processes.

    Notes
    -----
    Shared storage is organized using ``multiprocessing.Manager.dict``. This is separate
    process which starts work in the class constructor.
    """

    __instance = None

    def __init__(self):
        if ObjectStore.__instance is None:
            self.store_delayed = Manager().dict()

    def __repr__(self):
        return f"Object store: {self.store_delayed}"

    @classmethod
    def get_instance(cls):
        """
        Get instance of ``ObjectStore``.

        Returns
        -------
        unidist.core.backends.pymp.core.object_store.ObjectStore
        """
        if cls.__instance is None:
            cls.__instance = ObjectStore()
        return cls.__instance

    def put(self, data, data_id=None):
        """
        Put `data` to internal shared dictionary.

        Parameters
        ----------
        data : object
            Data to be put.
        data_id : unidist.core.backends.common.data_id.DataID, optional
            An ID to data. If it isn't provided, will be created automatically.

        Returns
        -------
        unidist.core.backends.common.data_id.DataID
            An ID of object in internal shared dictionary.
        """
        data_id = DataID() if data_id is None else data_id

        self.store_delayed[data_id] = pkl.dumps(data) if callable(data) else data
        return data_id

    def get(self, data_ids):
        """
        Get a object(s) associated with `data_ids` from the shared internal dictionary.

        Parameters
        ----------
        data_ids : unidist.core.backends.common.data_id.DataID or list
            An ID(s) of object(s) to get data from.

        Returns
        -------
        object
            A Python object.
        """
        is_list = isinstance(data_ids, list)
        if not is_list:
            data_ids = [data_ids]
        if not all(isinstance(ref, DataID) for ref in data_ids):
            raise ValueError(
                "`data_ids` must either be a data ID or a list of data IDs."
            )

        values = [None] * len(data_ids)

        for idx, data_id in enumerate(data_ids):
            while isinstance(self.store_delayed[data_id], Delayed):
                pass

            value = self.store_delayed[data_id]
            if isinstance(value, Exception):
                raise value

            values[idx] = pkl.loads(value) if isinstance(value, bytes) else value

        return values if is_list else values[0]

    def wait(self, data_ids, num_returns=1):
        """
        Wait until `data_ids` are finished.

        This method returns two lists. The first list consists of
        ``DataID``-s that correspond to objects that completed computations.
        The second list corresponds to the rest of the ``DataID``-s (which may or may not be ready).

        Parameters
        ----------
        data_ids : unidist.core.backends.common.data_id.DataID or list
            ``DataID`` or list of ``DataID``-s to be waited.
        num_returns : int, default: 1
            The number of ``DataID``-s that should be returned as ready.

        Returns
        -------
        tuple
            List of data IDs that are ready and list of the remaining data IDs.
        """
        if not isinstance(data_ids, list):
            data_ids = [data_ids]

        ready = list()
        not_ready = list()

        for idx, data_id in enumerate(data_ids[:]):
            if not isinstance(self.store_delayed[data_id], Delayed):
                ready.append(data_ids.pop(idx))

            if len(ready) == num_returns:
                break

        not_ready = data_ids

        while len(ready) != num_returns:
            self.get(not_ready[0])
            ready.append(not_ready.pop(0))

        return ready, not_ready



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""`ObjectStore` functionality."""

from unidist.core.backends.mpi.core.local_object_store import LocalObjectStore
from unidist.core.backends.mpi.core.serialization import deserialize_complex_data
from unidist.core.backends.mpi.core.shared_object_store import SharedObjectStore


class ObjectStore:
    """
    Class that combines checking and retrieving data from the shared and local stores in a current process.

    Notes
    -----
    The store checks for both deserialized and serialized data.
    """

    __instance = None

    @classmethod
    def get_instance(cls):
        """
        Get instance of ``ObjectStore``.

        Returns
        -------
        ObjectStore
        """
        if cls.__instance is None:
            cls.__instance = ObjectStore()
        return cls.__instance

    def contains(self, data_id):
        """
        Check if the data associated with `data_id` exists in the current process.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.

        Returns
        -------
        bool
            Return the status if an object exist in the current process.
        """
        local_store = LocalObjectStore.get_instance()
        shared_store = SharedObjectStore.get_instance()
        return (
            local_store.contains(data_id)
            or local_store.is_already_serialized(data_id)
            or shared_store.contains(data_id)
        )

    def get(self, data_id):
        """
        Get data from any location in the current process.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.

        Returns
        -------
        object
            Return data associated with `data_id`.
        """
        local_store = LocalObjectStore.get_instance()
        shared_store = SharedObjectStore.get_instance()

        if local_store.contains(data_id):
            return local_store.get(data_id)

        if local_store.is_already_serialized(data_id):
            serialized_data = local_store.get_serialized_data(data_id)
            value = deserialize_complex_data(
                serialized_data["s_data"],
                serialized_data["raw_buffers"],
                serialized_data["buffer_count"],
            )
        elif shared_store.contains(data_id):
            value = shared_store.get(data_id)
        else:
            raise ValueError("The current data ID is not contained in the procces.")
        local_store.put(data_id, value)
        return value



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""High-level API of Python Multiprocessing backend."""

import cloudpickle as pkl

from unidist.config import CpuCount
from unidist.core.backends.pymp.core.object_store import ObjectStore, Delayed
from unidist.core.backends.pymp.core.process_manager import (
    ProcessManager,
    Task,
)


# The global variable is responsible for if Python Multiprocessing backend has already been initialized
is_multiprocessing_initialized = False


def init(num_workers=CpuCount.get()):
    """
    Initialize shared object storage and workers pool.

    Parameters
    ----------
    num_workers : int, default: number of CPUs
        Number of worker-processes to start.

    Notes
    -----
    Run initialization of singleton objects ``unidist.core.backends.pymp.core.object_store.ObjectStore``
    and ``unidist.core.backends.pymp.core.process_manager.ProcessManager``.
    """
    ObjectStore.get_instance()
    ProcessManager.get_instance(num_workers=num_workers)
    global is_multiprocessing_initialized
    if not is_multiprocessing_initialized:
        is_multiprocessing_initialized = True


def is_initialized():
    """
    Check if Python Multiprocessing backend has already been initialized.

    Returns
    -------
    bool
        True or False.
    """
    global is_multiprocessing_initialized
    return is_multiprocessing_initialized


def put(data):
    """
    Put data into shared object storage.

    Parameters
    ----------
    data : object
        Data to be put.

    Returns
    -------
    unidist.core.backends.common.data_id.DataID
        An ID of object in shared object storage.
    """
    return ObjectStore.get_instance().put(data)


def get(data_ids):
    """
    Get a object(s) associated with `data_ids` from the shared object storage.

    Parameters
    ----------
    data_ids : unidist.core.backends.common.data_id.DataID or list
        An ID(s) to object(s) to get data from.

    Returns
    -------
    object
        A Python object.
    """
    return ObjectStore.get_instance().get(data_ids)


def wait(data_ids, num_returns=1):
    """
    Wait until `data_ids` are finished.

    This method returns two lists. The first list consists of
    ``DataID``-s that correspond to objects that completed computations.
    The second list corresponds to the rest of the ``DataID``-s (which may or may not be ready).

    Parameters
    ----------
    data_ids : unidist.core.backends.common.data_id.DataID or list
        ``DataID`` or list of ``DataID``-s to be waited.
    num_returns : int, default: 1
        The number of ``DataID``-s that should be returned as ready.

    Returns
    -------
    tuple
        List of data IDs that are ready and list of the remaining data IDs.
    """
    return ObjectStore.get_instance().wait(data_ids, num_returns=num_returns)


def submit(func, *args, num_returns=1, **kwargs):
    """
    Execute function in a worker process.

    Parameters
    ----------
    func : callable
        Function to be executed in the worker.
    *args : iterable
        Positional arguments to be passed in the `func`.
    num_returns : int, default: 1
        Number of results to be returned from `func`.
    **kwargs : dict
        Keyword arguments to be passed in the `func`.

    Returns
    -------
    unidist.core.backends.common.data_id.DataID, list or None
        Type of returns depends on `num_returns` value:

        * if `num_returns == 1`, ``DataID`` will be returned.
        * if `num_returns > 1`, list of ``DataID``-s will be returned.
        * if `num_returns == 0`, ``None`` will be returned.
    """
    obj_store = ObjectStore.get_instance()

    if num_returns == 0:
        data_ids = None
    elif num_returns > 1:
        data_ids = [obj_store.put(Delayed()) for _ in range(num_returns)]
    else:
        data_ids = obj_store.put(Delayed())

    task = Task(func, data_ids, obj_store, *args, **kwargs)

    ProcessManager.get_instance().submit(pkl.dumps(task))

    return data_ids



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""High-level API and utilities to implement it."""

import inspect


from unidist.core.base.utils import init_backend, get_backend_proxy


# --------------------------------------------------------------------------------------
# unidist decorator internal layer
# --------------------------------------------------------------------------------------
def make_decorator(num_cpus=None, num_returns=None, resources=None):
    """
    Make decorator for a user-provided function or a class.

    Parameters
    ----------
    num_cpus : int, optional
        The number of CPUs to reserve for the remote function or for the lifetime of the actor.
    num_returns : int, optional
        The number of ``ObjectRef``-s returned by the remote function invocation.
    resources : dict, optional
        Custom resources to reserve for the remote function or for the lifetime of the actor.

    Returns
    -------
    callable
        Function-decorator.
    """
    execution_backend = get_backend_proxy()

    def decorator(function_or_class):
        """
        Decorate a user-provided function or a class.

        Parameters
        ----------
        function_or_class : object
            Function or class to be decorated.

        Returns
        -------
        RemoteFunction or ActorClass
            Decorated function or decorated class.
        """
        if inspect.isfunction(function_or_class):
            if num_returns is not None and (
                not isinstance(num_returns, int) or num_returns < 0
            ):
                raise ValueError(
                    "The keyword 'num_returns' only accepts 0 or a positive integer"
                )

            return execution_backend.make_remote_function(
                function_or_class,
                num_cpus=num_cpus,
                num_returns=num_returns,
                resources=resources,
            )
        elif inspect.isclass(function_or_class):
            if num_returns is not None:
                raise TypeError("The keyword 'num_returns' is not allowed for actors.")

            return execution_backend.make_actor(
                function_or_class, num_cpus=num_cpus, resources=resources
            )
        else:
            raise TypeError(
                "The @unidist.remote decorator must be applied to either a function or to a class."
            )

    return decorator


# --------------------------------------------------------------------------------------
# unidist execution layer public API
# --------------------------------------------------------------------------------------
def init():
    """
    Initialize an execution backend.

    Notes
    -----
    The concrete execution backend can be set via
    `UNIDIST_BACKEND` environment variable or ``Backend`` config value.
    MPI backend is used by default.
    """
    init_backend()


def is_initialized():
    """
    Check if a unidist backend has already been initialized.

    Returns
    -------
    bool
        True or False.
    """
    execution_backend = get_backend_proxy()
    return execution_backend.is_initialized()


def shutdown():
    """Shutdown an execution backend."""
    execution_backend = get_backend_proxy()
    execution_backend.shutdown()


def remote(*args, **kwargs):
    """
    Define a remote function or an actor class.

    Parameters
    ----------
    *args : iterable
        Positional arguments to be passed in a remote function or an actor class.
    **kwargs : dict
        Keyword arguments to be passed in a remote function or an actor class.

    Returns
    -------
    RemoteFunction or ActorClass
    """
    # This is the case where the decorator is just @unidist.remote.
    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
        return make_decorator()(args[0])

    # Handle keyword arguments
    assert len(args) == 0 and len(kwargs) > 0, "Error"

    num_cpus = kwargs.get("num_cpus")
    num_returns = kwargs.get("num_returns")
    resources = kwargs.get("resources")
    return make_decorator(
        num_cpus=num_cpus, num_returns=num_returns, resources=resources
    )


def get(object_refs):
    """
    Get a remote object or a list of remote objects
    from distributed memory.

    Parameters
    ----------
    object_refs : ObjectRef or list
        ``ObjectRef`` or a list of ``ObjectRef``-s to get data from.

    Returns
    -------
    object
        A Python object or a list of Python objects.
    """
    execution_backend = get_backend_proxy()
    return execution_backend.get(object_refs)


def put(data):
    """
    Put data into distributed memory.

    Parameters
    ----------
    data : object
        Data to be put.

    Returns
    -------
    ObjectRef
        ``ObjectRef`` matching to data.
    """
    if not is_initialized():
        init()
    execution_backend = get_backend_proxy()
    return execution_backend.put(data)


def wait(object_refs, num_returns=1):
    """
    Wait until `object_refs` are finished.

    This method returns two lists. The first list consists of
    ``ObjectRef``-s that correspond to objects that completed computations.
    The second list corresponds to the rest of the ``ObjectRef``-s (which may or may not be ready).

    Parameters
    ----------
    object_refs : ObjectRef or list
        ``ObjectRef`` or list of ``ObjectRef``-s to be waited.
    num_returns : int, default: 1
        The number of ``ObjectRef``-s that should be returned as ready.

    Returns
    -------
    two lists
        List of ``ObjectRef``-s that are ready and list of the remaining ``ObjectRef``-s.
    """
    execution_backend = get_backend_proxy()
    return execution_backend.wait(object_refs, num_returns=num_returns)


def is_object_ref(obj):
    """
    Whether an object is ``ObjectRef`` or not.

    Parameters
    ----------
    obj : object
        An object to be checked.

    Returns
    -------
    bool
        `True` if an object is ``ObjectRef``, `False` otherwise.
    """
    execution_backend = get_backend_proxy()
    return execution_backend.is_object_ref(obj)


def get_ip():
    """
    Get node IP address.

    Returns
    -------
    str
        Node IP address.
    """
    execution_backend = get_backend_proxy()
    return execution_backend.get_ip()


def num_cpus():
    """
    Get the number of CPUs used by the execution backend.

    Returns
    -------
    int
    """
    execution_backend = get_backend_proxy()
    return execution_backend.num_cpus()


def cluster_resources():
    """
    Get resources of the cluster.

    Returns
    -------
    dict
        Dictionary with cluster nodes info in the form
        `{"node_ip0": {"CPU": x0}, "node_ip1": {"CPU": x1}, ...}`.
    """
    execution_backend = get_backend_proxy()
    return execution_backend.cluster_resources()



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Config entities specific for Ray backend which can be used for unidist behavior tuning."""

from .envvars import (
    RayGpuCount,
    IsRayCluster,
    RayRedisAddress,
    RayRedisPassword,
    RayObjectStoreMemory,
)

__all__ = [
    "RayGpuCount",
    "IsRayCluster",
    "RayRedisAddress",
    "RayRedisPassword",
    "RayObjectStoreMemory",
]



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Actor specific functionality using MPI backend."""

import unidist.core.backends.mpi.core.common as common
import unidist.core.backends.mpi.core.communication as communication
from unidist.core.backends.mpi.core.async_operations import AsyncOperations
from unidist.core.backends.mpi.core.local_object_store import LocalObjectStore
from unidist.core.backends.mpi.core.controller.garbage_collector import (
    garbage_collector,
)
from unidist.core.backends.mpi.core.controller.common import push_data, RoundRobin
from unidist.core.backends.mpi.core.controller.api import put


class ActorMethod:
    """
    Class responsible to execute method of an actor.

    Execute `method_name` of `actor` in a separate worker process, where ``Actor`` object is created.

    Parameters
    ----------
    actor : unidist.core.backends.mpi.core.executor.Actor
        Actor object.
    method_name : str
        The name of the method to be called.
    """

    def __init__(self, actor, method_name):
        self._actor = actor
        self._method_name = method_name

    def __call__(self, *args, num_returns=1, **kwargs):
        local_store = LocalObjectStore.get_instance()
        output_id = local_store.generate_output_data_id(
            self._actor._owner_rank, garbage_collector, num_returns
        )

        push_data(self._actor._owner_rank, self._method_name)
        push_data(self._actor._owner_rank, args)
        push_data(self._actor._owner_rank, kwargs)

        operation_type = common.Operation.ACTOR_EXECUTE
        operation_data = {
            "task": self._method_name,
            # tuple cannot be serialized iteratively and it will fail if some internal data cannot be serialized using Pickle
            "args": list(args),
            "kwargs": kwargs,
            "output": output_id,
            "handler": self._actor._handler_id,
        }
        async_operations = AsyncOperations.get_instance()
        h_list, _ = communication.isend_complex_operation(
            communication.MPIState.get_instance().global_comm,
            operation_type,
            operation_data,
            self._actor._owner_rank,
            is_serialized=False,
        )
        async_operations.extend(h_list)
        return output_id


class Actor:
    """
    Class to execute methods of a wrapped class in a separate worker process.

    Parameters
    ----------
    cls : object
        Class to be an actor class.
    *args : iterable
        Positional arguments to be passed in `cls` constructor.
    owner_rank : None or int
        MPI rank which an actor class should be created on.
        Used for proper serialization/deserialization via `__reduce__`.
    handler_id : None or unidist.core.backends.mpi.core.common.MpiDataID
        An ID to the actor class.
        Used for proper serialization/deserialization via `__reduce__`.
    **kwargs : dict
        Keyword arguments to be passed in `cls` constructor.

    Notes
    -----
    Instance of the `cls` will be created on the worker.
    """

    def __init__(self, cls, *args, owner_rank=None, handler_id=None, **kwargs):
        self._cls = cls
        self._args = args
        self._kwargs = kwargs
        self._owner_rank = (
            RoundRobin.get_instance().schedule_rank()
            if owner_rank is None
            else owner_rank
        )
        local_store = LocalObjectStore.get_instance()
        self._handler_id = (
            local_store.generate_data_id(garbage_collector)
            if handler_id is None
            else handler_id
        )
        local_store.put_data_owner(self._handler_id, self._owner_rank)

        # reserve a rank for actor execution only
        RoundRobin.get_instance().reserve_rank(self._owner_rank)

        # submit `ACTOR_CREATE` task to a worker only once
        if owner_rank is None and handler_id is None:
            operation_type = common.Operation.ACTOR_CREATE
            operation_data = {
                "class": cls,
                "args": args,
                "kwargs": kwargs,
                "handler": self._handler_id,
            }
            async_operations = AsyncOperations.get_instance()
            h_list, _ = communication.isend_complex_operation(
                communication.MPIState.get_instance().global_comm,
                operation_type,
                operation_data,
                self._owner_rank,
                is_serialized=False,
            )
            async_operations.extend(h_list)

    def _serialization_helper(self):
        """
        Helper to save the state of the object.

        This is defined to make pickling work via `__reduce__`.

        Returns
        -------
        dict
            A dictionary of the information needed to reconstruct the object.
        """
        state = {
            "cls": self._cls,
            "args": self._args,
            "owner_rank": self._owner_rank,
            "handler_id": self._handler_id,
            "kwargs": self._kwargs,
        }
        return state

    @classmethod
    def _deserialization_helper(cls, state):
        """
        Helper to restore the state of the object.

        This is defined to make pickling work via `__reduce__`.

        Parameters
        ----------
        state : dict
            The serialized state of the object.

        Returns
        -------
        Actor
        """
        return cls(
            state["cls"],
            *state["args"],
            owner_rank=state["owner_rank"],
            handler_id=state["handler_id"],
            **state["kwargs"],
        )

    def __reduce__(self):
        """
        This is defined intentionally to make pickling work correctly.

        Returns
        -------
        tuple
            Callable and arguments to be passed in to it.
        """
        state = self._serialization_helper()
        return self._deserialization_helper, (state,)

    # Cache for serialized actor methods {"method_name": DataID}
    actor_methods = {}

    def __getattr__(self, name):
        data_id_to_method = self.actor_methods.get(name, None)
        if data_id_to_method is None:
            data_id_to_method = put(name)
            self.actor_methods[name] = data_id_to_method
        return ActorMethod(self, data_id_to_method)

    def __del__(self):
        """
        This is defined to release the rank reserved for the actor when it gets out of scope.
        """
        RoundRobin.get_instance().release_rank(self._owner_rank)



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""`LocalObjectStore` functionality."""

import weakref

import unidist.core.backends.mpi.core.common as common
import unidist.core.backends.mpi.core.communication as communication


class LocalObjectStore:
    """
    Class that stores local objects and provides access to them.

    Notes
    -----
    The storage is local to the current worker process only.
    """

    __instance = None

    def __init__(self):
        # Add local data {DataID : Data}
        self._data_map = weakref.WeakKeyDictionary()
        # "strong" references to data IDs {DataID : DataID}
        # we are using dict here to improve performance when getting an element from it,
        # whereas other containers would require O(n) complexity
        self._data_id_map = {}
        # Data owner {DataID : Rank}
        self._data_owner_map = weakref.WeakKeyDictionary()
        # Data was already sent to this ranks {DataID : [ranks]}
        self._sent_data_map = weakref.WeakKeyDictionary()
        # Data id generator
        self._data_id_counter = 0
        # Data serialized cache
        self._serialization_cache = weakref.WeakKeyDictionary()

    @classmethod
    def get_instance(cls):
        """
        Get instance of ``LocalObjectStore``.

        Returns
        -------
        LocalObjectStore
        """
        if cls.__instance is None:
            cls.__instance = LocalObjectStore()
        return cls.__instance

    def maybe_update_data_id_map(self, data_id):
        """
        Add a strong reference to the `data_id` if necessary.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.

        Notes
        -----
        The worker must have a strong reference to the external `data_id` until the owner process
        send the `unidist.core.backends.common.Operation.CLEANUP` operation with this `data_id`.
        """
        if (
            data_id.owner_rank != communication.MPIState.get_instance().global_rank
            and data_id not in self._data_id_map
        ):
            self._data_id_map[data_id] = data_id

    def put(self, data_id, data):
        """
        Put `data` to internal dictionary.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.
        data : object
            Data to be put.
        """
        self._data_map[data_id] = data
        self.maybe_update_data_id_map(data_id)

    def put_data_owner(self, data_id, rank):
        """
        Put data location (owner rank) to internal dictionary.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.
        rank : int
            Rank number where the data resides.
        """
        self._data_owner_map[data_id] = rank
        self.maybe_update_data_id_map(data_id)

    def get(self, data_id):
        """
        Get the data from a local dictionary.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.

        Returns
        -------
        object
            Return local data associated with `data_id`.
        """
        return self._data_map[data_id]

    def get_data_owner(self, data_id):
        """
        Get the data owner rank.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.

        Returns
        -------
        int
            Rank number where the data resides.
        """
        return self._data_owner_map[data_id]

    def contains(self, data_id):
        """
        Check if the data associated with `data_id` exists in a local dictionary.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.

        Returns
        -------
        bool
            Return the status if an object exist in local dictionary.
        """
        return data_id in self._data_map

    def contains_data_owner(self, data_id):
        """
        Check if the data location info associated with `data_id` exists in a local dictionary.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.

        Returns
        -------
        bool
            Return the ``True`` status if an object location is known.
        """
        return data_id in self._data_owner_map

    def clear(self, cleanup_list):
        """
        Clear "strong" references to data IDs from `cleanup_list`.

        Parameters
        ----------
        cleanup_list : list
            List of data IDs.

        Notes
        -----
        The actual data will be collected later when there is no weak or
        strong reference to data in the current worker.
        """
        for data_id in cleanup_list:
            self._data_id_map.pop(data_id, None)

    def generate_data_id(self, gc):
        """
        Generate unique ``MpiDataID`` instance.

        Parameters
        ----------
        gc : unidist.core.backends.mpi.core.executor.GarbageCollector
            Local garbage collector reference.

        Returns
        -------
        unidist.core.backends.mpi.core.common.MpiDataID
            Unique data ID instance.
        """
        data_id = common.MpiDataID(
            communication.MPIState.get_instance().global_rank, self._data_id_counter, gc
        )
        self._data_id_counter += 1
        return data_id

    def generate_output_data_id(self, dest_rank, gc, num_returns=1):
        """
        Generate unique list of ``unidist.core.backends.mpi.core.common.MpiDataID`` instance.

        Parameters
        ----------
        dest_rank : int
            Ranks number where generated list will be located.
        gc : unidist.core.backends.mpi.core.executor.GarbageCollector
            Local garbage collector reference.
        num_returns : int, default: 1
            Generated list size.

        Returns
        -------
        list
            A list of unique ``MpiDataID`` instances.
        """
        if num_returns == 1:
            output_ids = self.generate_data_id(gc)
            self.put_data_owner(output_ids, dest_rank)
        elif num_returns == 0:
            output_ids = None
        else:
            output_ids = [None] * num_returns
            for i in range(num_returns):
                output_id = self.generate_data_id(gc)
                output_ids[i] = output_id
                self.put_data_owner(output_id, dest_rank)
        return output_ids

    def cache_send_info(self, data_id, rank):
        """
        Save communication event for this `data_id` and rank.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ``ID`` to data.
        rank : int
            Rank number where the data was sent.
        """
        if data_id in self._sent_data_map:
            self._sent_data_map[data_id].add(rank)
        else:
            self._sent_data_map[data_id] = set([rank])

    def is_already_sent(self, data_id, rank):
        """
        Check if communication event on this `data_id` and rank happened.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data
        rank : int
            Rank number to check.

        Returns
        -------
        bool
            ``True`` if communication event already happened.
        """
        return (data_id in self._sent_data_map) and (
            rank in self._sent_data_map[data_id]
        )

    def cache_serialized_data(self, data_id, data):
        """
        Save serialized object for this `data_id`.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.
        data : object
            Serialized data to cache.
        """
        # We make a copy to avoid data corruption obtained through out-of-band serialization,
        # and buffers are marked read-only to prevent them from being modified.
        # `to_bytes()` call handles both points.
        data["raw_buffers"] = [buf.tobytes() for buf in data["raw_buffers"]]
        self._serialization_cache[data_id] = data
        self.maybe_update_data_id_map(data_id)

    def is_already_serialized(self, data_id):
        """
        Check if the data on this `data_id` is already serialized.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.

        Returns
        -------
        bool
            ``True`` if the data is already serialized.
        """
        return data_id in self._serialization_cache

    def get_serialized_data(self, data_id):
        """
        Get serialized data on this `data_id`.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.

        Returns
        -------
        object
            Cached serialized data associated with `data_id`.
        """
        return self._serialization_cache[data_id]



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Config entities which can be used for unidist common behavior tuning."""

from packaging import version

from unidist.config.parameter import EnvironmentVariable
from unidist.core.base.common import BackendName


class Backend(EnvironmentVariable, type=str):
    """Distribution backend to run queries by."""

    varname = "UNIDIST_BACKEND"
    choices = (
        BackendName.RAY,
        BackendName.DASK,
        BackendName.MPI,
        BackendName.PYMP,
        BackendName.PYSEQ,
    )

    @classmethod
    def _get_default(cls):
        """
        Get default value of the config.

        Returns
        -------
        str
        """
        try:
            import mpi4py
        except ImportError:
            pass
        else:
            if version.parse(mpi4py.__version__) < version.parse("3.0.3"):
                raise ImportError(
                    "Please `pip install unidist[mpi]` to install compatible MPI version."
                )
            return BackendName.MPI
        try:
            import ray

        except ImportError:
            pass
        else:
            if version.parse(ray.__version__) < version.parse("1.4.0"):
                raise ImportError(
                    "Please `pip install unidist[ray]` to install compatible Ray version."
                )
            return BackendName.RAY
        try:
            import dask
            import distributed

        except ImportError:
            pass
        else:
            if version.parse(dask.__version__) < version.parse(
                "2.22.0"
            ) or version.parse(distributed.__version__) < version.parse("2.22.0"):
                raise ImportError(
                    "Please `pip install unidist[dask]` to install compatible Dask version."
                )
            return BackendName.DASK

        return BackendName.PYMP


class CpuCount(EnvironmentVariable, type=int):
    """How many CPU cores to use during initialization of the unidist backend."""

    varname = "UNIDIST_CPUS"

    @classmethod
    def _get_default(cls):
        """
        Get default value of the config.

        Returns
        -------
        int
        """
        import multiprocessing

        return multiprocessing.cpu_count()



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""An implementation of ``Backend`` interface using Python Sequential backend."""

import socket

import unidist.core.backends.pyseq.core as py
from unidist.core.backends.pyseq.actor import PySeqActor
from unidist.core.backends.pyseq.remote_function import (
    PySeqRemoteFunction,
)
from unidist.core.base.backend import Backend


class PySeqBackend(Backend):
    """The class that implements the interface in ``Backend`` using Python Sequential backend."""

    @staticmethod
    def make_remote_function(function, num_cpus, num_returns, resources):
        """
        Define ``PySeqRemoteFunction``.

        function : callable
            Function to be ``PySeqRemoteFunction``.
        num_cpus : int
            The number of CPUs to reserve for ``PySeqRemoteFunction``.
        num_returns : int
            The number of ``ObjectRef``-s returned by the function invocation.
        resources : dict
            Custom resources to reserve for the function.

        Returns
        -------
        PySeqRemoteFunction
        """
        return PySeqRemoteFunction(function, num_cpus, num_returns, resources)

    @staticmethod
    def make_actor(cls, num_cpus, resources):
        """
        Define an actor class.

        cls : object
            Class to be an actor class.
        num_cpus : int
            The number of CPUs to reserve for the lifetime of the actor.
        resources : dict
            Custom resources to reserve for the lifetime of the actor.

        Returns
        -------
        PySeqActor
            The actor class type to create.
        list
            The list of arguments for ``PySeqActor`` constructor.
        """
        return PySeqActor, [cls, num_cpus, resources]

    @staticmethod
    def get(data_ids):
        """
        Get an object or a list of objects from object store.

        Parameters
        ----------
        data_ids : unidist.core.backends.common.data_id.DataID or list
            ``DataID`` or a list of ``DataID`` objects to get data from.

        Returns
        -------
        object
            A Python object or a list of Python objects.
        """
        return py.get(data_ids)

    @staticmethod
    def put(data):
        """
        Put `data` into object store.

        Parameters
        ----------
        data : object
            Data to be put.

        Returns
        -------
        unidist.core.backends.common.data_id.DataID
            ``DataID`` matching to data.
        """
        return py.put(data)

    @staticmethod
    def wait(data_ids, num_returns=1):
        """
        Wait until `data_ids` are finished.

        This method returns two lists. The first list consists of
        data IDs that correspond to objects that completed computations.
        The second list corresponds to the rest of the data IDs.

        Parameters
        ----------
        object_refs : unidist.core.backends.common.data_id.DataID or list
            ``DataID`` or list of ``DataID``-s to be waited.
        num_returns : int, default: 1
            The number of ``DataID``-s that should be returned as ready.

        Returns
        -------
        tuple
            List of data IDs that are ready and list of the remaining data IDs.

        Notes
        -----
        Method serves to maintain behavior compatibility between backends. All objects
        completed computation before putting into an object storage for Python Sequential backend.
        """
        return data_ids[:num_returns], data_ids[num_returns:]

    @staticmethod
    def get_ip():
        """
        Get node IP address.

        Returns
        -------
        str
            Node IP address.
        """
        hostname = socket.gethostname()
        return socket.gethostbyname(hostname)

    @staticmethod
    def num_cpus():
        """
        Get the number of CPUs used by the execution backend.

        Returns
        -------
        int
        """
        return 1

    @staticmethod
    def cluster_resources():
        """
        Get resources of the cluster.

        Returns
        -------
        dict
            Dictionary with node info in the form `{"node_ip": {"CPU": x}}`.
        """
        return {PySeqBackend.get_ip(): {"CPU": PySeqBackend.num_cpus()}}

    @staticmethod
    def is_initialized():
        """
        Check if Python Sequential backend has already been initialized.

        Returns
        -------
        bool
            True or False.
        """
        return py.is_initialized()



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Common functionality related to `controller`."""

import itertools

from unidist.core.backends.common.data_id import is_data_id
import unidist.core.backends.mpi.core.common as common
import unidist.core.backends.mpi.core.communication as communication
from unidist.core.backends.mpi.core.async_operations import AsyncOperations
from unidist.core.backends.mpi.core.local_object_store import LocalObjectStore
from unidist.core.backends.mpi.core.shared_object_store import SharedObjectStore
from unidist.core.backends.mpi.core.serialization import (
    serialize_complex_data,
)


logger = common.get_logger("common", "common.log")


class RoundRobin:
    __instance = None

    def __init__(self):
        self.reserved_ranks = []
        mpi_state = communication.MPIState.get_instance()
        self.rank_to_schedule = itertools.cycle(
            (
                global_rank
                for global_rank in mpi_state.workers
                # check if a rank to schedule is not equal to the rank
                # of the current process to not get into recursive scheduling
                if global_rank != mpi_state.global_rank
            )
        )
        logger.debug(f"RoundRobin init for {mpi_state.global_rank} rank")

    @classmethod
    def get_instance(cls):
        """
        Get instance of ``RoundRobin``.

        Returns
        -------
        RoundRobin
        """
        if cls.__instance is None:
            cls.__instance = RoundRobin()
        return cls.__instance

    def schedule_rank(self):
        """
        Find the next non-reserved rank for task/actor-task execution.

        Returns
        -------
        int
            A rank number.
        """
        next_rank = None
        mpi_state = communication.MPIState.get_instance()

        # Go rank by rank to find the first one non-reserved
        for _ in mpi_state.workers:
            rank = next(self.rank_to_schedule)
            if rank not in self.reserved_ranks:
                next_rank = rank
                break

        if next_rank is None:
            raise Exception("All ranks blocked")

        return next_rank

    def reserve_rank(self, rank):
        """
        Reserve the rank for the actor scheduling.

        This makes the rank unavailable for scheduling a new actor or task
        until it gets released.

        Parameters
        ----------
        rank : int
            A rank number.
        """
        self.reserved_ranks.append(rank)
        logger.debug(
            f"RoundRobin reserve rank {rank} for actor "
            + f"on worker with rank {communication.MPIState.get_instance().global_rank}"
        )

    def release_rank(self, rank):
        """
        Release the rank reserved for the actor.

        This makes the rank available for scheduling a new actor or task.

        Parameters
        ----------
        rank : int
            A rank number.
        """
        self.reserved_ranks.remove(rank)
        logger.debug(
            f"RoundRobin release rank {rank} reserved for actor "
            + f"on worker with rank {communication.MPIState.get_instance().global_rank}"
        )


def pull_data(comm, owner_rank=None):
    """
    Receive data from another MPI process.

    Data can come from shared memory or direct communication depending on the package type.
    The data is de-serialized from received buffers.

    Parameters
    ----------
    owner_rank : int or None
        Source MPI process to receive data from.
        If ``None``, data will be received from any source based on `iprobe`.

    Returns
    -------
    object
        Received data object from another MPI process.
    """
    if owner_rank is None:
        info_package, source = communication.mpi_iprobe_recv_object(
            comm, tag=common.MPITag.OBJECT_BLOCKING
        )
        owner_rank = source
    else:
        info_package = communication.mpi_recv_object(comm, owner_rank)
    if info_package["package_type"] == common.MetadataPackage.SHARED_DATA:
        local_store = LocalObjectStore.get_instance()
        shared_store = SharedObjectStore.get_instance()
        data_id = info_package["id"]

        if local_store.contains(data_id):
            return {
                "id": data_id,
                "data": local_store.get(data_id),
            }

        data = shared_store.get(data_id, owner_rank, info_package)
        local_store.put(data_id, data)
        return {
            "id": data_id,
            "data": data,
        }
    elif info_package["package_type"] == common.MetadataPackage.LOCAL_DATA:
        local_store = LocalObjectStore.get_instance()
        data = communication.recv_complex_data(
            comm, owner_rank, info_package=info_package
        )
        data_id = info_package["id"]
        local_store.put(data_id, data)
        return {
            "id": data_id,
            "data": data,
        }
    elif info_package["package_type"] == common.MetadataPackage.TASK_DATA:
        return communication.recv_complex_data(
            comm, owner_rank, info_package=info_package
        )
    else:
        raise ValueError("Unexpected package of data info!")


def request_worker_data(data_ids):
    """
    Get objects associated with `data_ids` from the object storage.

    Parameters
    ----------
    data_ids : list[unidist.core.backends.mpi.core.common.MpiDataID]
        IDs to objects to get data from.

    Returns
    -------
    object
        A Python object.
    """
    mpi_state = communication.MPIState.get_instance()
    local_store = LocalObjectStore.get_instance()
    async_operations = AsyncOperations.get_instance()

    for data_id in data_ids:
        owner_rank = local_store.get_data_owner(data_id)

        logger.debug("GET {} id from {} rank".format(data_id._id, owner_rank))

        # Worker request
        operation_type = common.Operation.GET
        operation_data = {
            "source": mpi_state.global_rank,
            "id": data_id,
            # set `is_blocking_op` to `True` to tell a worker
            # to send the data directly to the requester
            # without any delay
            "is_blocking_op": True,
        }
        h_list = communication.isend_simple_operation(
            mpi_state.global_comm,
            operation_type,
            operation_data,
            owner_rank,
        )
        # We do not wait for async requests here because
        # we can receive the data from the first available worker below
        async_operations.extend(h_list)

    data_count = 0
    # If some dataids raise an exception it will be captured in exception_raised variable and raised after the while loop ends.
    exception_raised = None
    while data_count < len(data_ids):
        # Remote data gets available in the local store inside `pull_data`
        complex_data = pull_data(mpi_state.global_comm)
        if isinstance(complex_data["data"], Exception) and exception_raised is None:
            exception_raised = complex_data["data"]
        data_id = complex_data["id"]
        if data_id in data_ids:
            data_count += 1
        else:
            raise RuntimeError(
                f"DataID {data_id} isn't in the requested list {data_ids}"
            )
    if exception_raised:
        raise exception_raised


def _push_local_data(dest_rank, data_id, is_blocking_op, is_serialized):
    """
    Send local data associated with passed ID to target rank.

    Parameters
    ----------
    dest_rank : int
        Target rank.
    data_id : unidist.core.backends.mpi.core.common.MpiDataID
        An ID to data.
    is_blocking_op : bool
        Whether the communication should be blocking or not.
        If ``True``, the request should be processed immediatly
        even for a worker since it can get into controller mode.
    is_serialized : bool
        `data_id` is already serialized or not.
    """
    local_store = LocalObjectStore.get_instance()

    # Check if data was already pushed
    if not local_store.is_already_sent(data_id, dest_rank):
        logger.debug("PUT LOCAL {} id to {} rank".format(data_id._id, dest_rank))

        mpi_state = communication.MPIState.get_instance()
        async_operations = AsyncOperations.get_instance()
        # Push the local master data to the target worker directly
        if is_serialized:
            operation_data = local_store.get_serialized_data(data_id)
            # Insert `data_id` to get full metadata package further
            operation_data["id"] = data_id
        else:
            operation_data = {
                "id": data_id,
                "data": local_store.get(data_id),
            }

        operation_type = common.Operation.PUT_DATA
        if is_blocking_op:
            serialized_data = communication.send_complex_data(
                mpi_state.global_comm,
                operation_data,
                dest_rank,
                is_serialized=is_serialized,
            )
        else:
            h_list, serialized_data = communication.isend_complex_operation(
                mpi_state.global_comm,
                operation_type,
                operation_data,
                dest_rank,
                is_serialized=is_serialized,
            )
            async_operations.extend(h_list)

        if not is_serialized or not local_store.is_already_serialized(data_id):
            local_store.cache_serialized_data(data_id, serialized_data)

        #  Remember pushed id
        local_store.cache_send_info(data_id, dest_rank)


def _push_shared_data(dest_rank, data_id, is_blocking_op):
    """
    Send the data associated with the data ID to target rank using shared memory.

    Parameters
    ----------
    dest_rank : int
        Target rank.
    data_id : unidist.core.backends.mpi.core.common.MpiDataID
        An ID to data.
    is_blocking_op : bool
        Whether the communication should be blocking or not.
        If ``True``, the request should be processed immediatly
        even for a worker since it can get into controller mode.
    """
    local_store = LocalObjectStore.get_instance()
    # Check if data was already pushed
    if not local_store.is_already_sent(data_id, dest_rank):
        mpi_state = communication.MPIState.get_instance()
        shared_store = SharedObjectStore.get_instance()
        mpi_state = communication.MPIState.get_instance()
        operation_type = common.Operation.PUT_SHARED_DATA
        async_operations = AsyncOperations.get_instance()
        info_package = shared_store.get_shared_info(data_id)
        # wrap to dict for sending and correct deserialization of the object by the recipient
        operation_data = dict(info_package)
        if is_blocking_op:
            communication.mpi_send_object(
                mpi_state.global_comm,
                operation_data,
                dest_rank,
                tag=common.MPITag.OBJECT_BLOCKING,
            )
        else:
            h_list = communication.isend_simple_operation(
                mpi_state.global_comm,
                operation_type,
                operation_data,
                dest_rank,
            )
            async_operations.extend(h_list)
        local_store.cache_send_info(data_id, dest_rank)


def _push_data_owner(dest_rank, data_id):
    """
    Send data location associated with data ID to target rank.

    Parameters
    ----------
    dest_rank : int
        Target rank.
    data_id : unidist.core.backends.mpi.core.common.MpiDataID
        An ID to data.
    """
    local_store = LocalObjectStore.get_instance()
    operation_type = common.Operation.PUT_OWNER
    operation_data = {
        "id": data_id,
        "owner": local_store.get_data_owner(data_id),
    }
    async_operations = AsyncOperations.get_instance()
    h_list = communication.isend_simple_operation(
        communication.MPIState.get_instance().global_comm,
        operation_type,
        operation_data,
        dest_rank,
    )
    async_operations.extend(h_list)


def push_data(dest_rank, value, is_blocking_op=False):
    """
    Parse and send all values to destination rank.

    Process all arguments recursivelly and send all ID associated data or it's location
    to the target rank.

    Parameters
    ----------
    dest_rank : int
        Rank where task data is needed.
    value : iterable or dict or object
        Arguments to be sent.
    is_blocking_op : bool
        Whether the communication should be blocking or not.
        If ``True``, the request should be processed immediatly
        even for a worker since it can get into controller mode.
    """
    local_store = LocalObjectStore.get_instance()
    shared_store = SharedObjectStore.get_instance()

    if isinstance(value, (list, tuple)):
        for v in value:
            push_data(dest_rank, v)
    elif isinstance(value, dict):
        for v in value.values():
            push_data(dest_rank, v)
    elif is_data_id(value):
        data_id = value
        if shared_store.contains(data_id):
            _push_shared_data(dest_rank, data_id, is_blocking_op)
        elif local_store.is_already_serialized(data_id):
            _push_local_data(dest_rank, data_id, is_blocking_op, is_serialized=True)
        elif local_store.contains(data_id):
            data = local_store.get(data_id)
            serialized_data = serialize_complex_data(data)
            if shared_store.is_allocated() and shared_store.should_be_shared(
                serialized_data
            ):
                shared_store.put(data_id, serialized_data)
                _push_shared_data(dest_rank, data_id, is_blocking_op)
            else:
                local_store.cache_serialized_data(data_id, serialized_data)
                _push_local_data(dest_rank, data_id, is_blocking_op, is_serialized=True)
        elif local_store.contains_data_owner(data_id):
            _push_data_owner(dest_rank, data_id)
        else:
            raise ValueError("Unknown DataID!")



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""MPI backend functionality related to `controller` concept."""

from unidist.core.backends.mpi.core.controller.api import (
    put,
    get,
    submit,
    wait,
    init,
    is_initialized,
    cluster_resources,
    shutdown,
)
from unidist.core.backends.mpi.core.controller.actor import Actor

__all__ = [
    "put",
    "get",
    "submit",
    "wait",
    "init",
    "is_initialized",
    "cluster_resources",
    "shutdown",
    "Actor",
]



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import sys
import pytest
import gc

import unidist
from unidist.config import Backend, CpuCount
from unidist.core.base.common import BackendName
from .utils import assert_equal, TestAsyncActor

unidist.init()


@pytest.fixture(autouse=True)
def call_gc_collect():
    """
    Collect all references from the previous test in order for MPI backend to work correctly.
    """
    yield
    # This is only needed for the MPI backend
    if Backend.get() == BackendName.MPI:
        gc.collect()


@pytest.mark.skipif(
    Backend.get() in (BackendName.PYMP, BackendName.PYSEQ),
    reason="`pymp` and `pyseq` backends do not support execution of coroutines.",
)
@pytest.mark.skipif(
    sys.platform == "win32" and Backend.get() == BackendName.PYMP,
    reason="Details are in https://github.com/modin-project/unidist/issues/70.",
)
@pytest.mark.parametrize("is_default_constructor", [True, False])
def test_actor_constructor(is_default_constructor):
    actor = (
        TestAsyncActor.remote() if is_default_constructor else TestAsyncActor.remote(5)
    )
    assert_equal(actor.get_accumulator.remote(), 0 if is_default_constructor else 5)


@pytest.mark.skipif(
    Backend.get() in (BackendName.PYMP, BackendName.PYSEQ),
    reason="`pymp` and `pyseq` backends do not support execution of coroutines.",
)
@pytest.mark.skipif(
    sys.platform == "win32" and Backend.get() == BackendName.PYMP,
    reason="Details are in https://github.com/modin-project/unidist/issues/70.",
)
def test_chaining():
    object_ref = unidist.put(7)
    actor = TestAsyncActor.remote()
    assert_equal(actor.task.remote(object_ref), 8)


@pytest.mark.skipif(
    Backend.get() in (BackendName.PYMP, BackendName.PYSEQ),
    reason="`pymp` and `pyseq` backends do not support execution of coroutines.",
)
@pytest.mark.skipif(
    sys.platform == "win32" and Backend.get() == BackendName.PYMP,
    reason="Details are in https://github.com/modin-project/unidist/issues/70.",
)
def test_num_returns():
    actor = TestAsyncActor.remote(7)
    object_ref0, object_ref1 = actor.multiple_returns.options(num_returns=2).remote(3)
    assert_equal(object_ref0, 3)
    assert_equal(object_ref1, 7)


@pytest.mark.skipif(
    Backend.get() in (BackendName.PYMP, BackendName.PYSEQ),
    reason="`pymp` and `pyseq` backends do not support execution of coroutines.",
)
@pytest.mark.skipif(
    Backend.get() == BackendName.PYMP,
    reason="`pymp` backend incorrectly frees grabbed actors. Details are in https://github.com/modin-project/unidist/issues/65.",
)
@pytest.mark.skipif(
    sys.platform == "win32" and Backend.get() == BackendName.PYMP,
    reason="Details are in https://github.com/modin-project/unidist/issues/70.",
)
@pytest.mark.parametrize("is_use_options", [True, False])
def test_address_space(is_use_options):
    actor0 = (
        TestAsyncActor.options().remote(0)
        if is_use_options
        else TestAsyncActor.remote(0)
    )
    actor1 = (
        TestAsyncActor.options().remote(1)
        if is_use_options
        else TestAsyncActor.remote(1)
    )

    object_ref0 = actor0.get_accumulator.remote()
    object_ref1 = actor1.get_accumulator.remote()

    assert_equal(object_ref0, 0)
    assert_equal(object_ref1, 1)


@pytest.mark.skipif(
    Backend.get() in (BackendName.PYMP, BackendName.PYSEQ),
    reason="Proper serialization/deserialization is not implemented yet for `pymp` and `pyseq`",
)
def test_global_capture():
    actor = TestAsyncActor.remote(0)

    @unidist.remote
    def foo():
        object_ref = actor.get_accumulator.remote()
        return unidist.get(object_ref)

    object_ref = foo.remote()

    assert_equal(object_ref, 0)


@pytest.mark.skipif(
    Backend.get() in (BackendName.PYMP, BackendName.PYSEQ),
    reason="Proper serialization/deserialization is not implemented yet for `pymp` and `pyseq`",
)
def test_direct_capture():
    actor = TestAsyncActor.remote(0)

    @unidist.remote
    def foo(actor):
        object_ref = actor.get_accumulator.remote()
        return unidist.get(object_ref)

    object_ref = foo.remote(actor)

    assert_equal(object_ref, 0)


@pytest.mark.skipif(
    Backend.get() in (BackendName.PYMP, BackendName.PYSEQ),
    reason="Details are in https://github.com/modin-project/unidist/issues/70.",
)
def test_return_none():
    actor = TestAsyncActor.remote()
    assert_equal(actor.task_return_none.remote(), None)


@pytest.mark.skipif(
    Backend.get() in (BackendName.PYMP, BackendName.PYSEQ),
    reason="`pymp` and `pyseq` backends do not support execution of coroutines.",
)
def test_pending_get():
    @unidist.remote
    class SlowActor:
        async def slow_execute(self):
            await asyncio.sleep(5)
            return 1

    slow_actor = SlowActor.remote()

    @unidist.remote
    def g():
        return unidist.get(slow_actor.slow_execute.remote()) + 1

    assert_equal(g.remote(), 2)


@pytest.mark.skipif(
    Backend.get() == BackendName.DASK,
    reason="Unexpected exception `There is no current event loop in thread` is raised",
)
@pytest.mark.skipif(
    Backend.get() == BackendName.PYMP,
    reason="Run of a remote task inside of an async actor method is not implemented yet for `pymp`",
)
def test_signal_actor():
    @unidist.remote
    class SignalActor:
        def __init__(self, event_count: int):
            self.events = [asyncio.Event() for _ in range(event_count)]

        def send(self, event_idx: int):
            self.events[event_idx].set()

        async def wait(self, event_idx: int):
            await self.events[event_idx].wait()

    signals = SignalActor.remote(CpuCount.get() + 1)

    unidist.get(signals.send.remote(0))

    @unidist.remote
    def func(idx):
        unidist.get(signals.wait.remote(idx))
        unidist.get(signals.send.remote(idx + 1))
        return idx

    object_refs = [func.remote(idx) for idx in range(CpuCount.get())]

    assert_equal(object_refs, list(range(CpuCount.get())))



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Config entities specific for Dask backend which can be used for unidist behavior tuning."""

from unidist.config.parameter import EnvironmentVariable, ExactStr


class DaskMemoryLimit(EnvironmentVariable, type=int):
    """How many bytes of memory that Dask worker should use."""

    varname = "UNIDIST_DASK_MEMORY_LIMIT"


class IsDaskCluster(EnvironmentVariable, type=bool):
    """Whether Dask is running on pre-initialized Dask cluster."""

    varname = "UNIDIST_DASK_CLUSTER"


class DaskSchedulerAddress(EnvironmentVariable, type=ExactStr):
    """Dask Scheduler address to connect to when running in Dask cluster."""

    varname = "UNIDIST_DASK_SCHEDULER_ADDRESS"



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Utilities used to initialize Python Sequential execution backend."""


def initialize_pyseq():
    """
    Initialize the Python Sequential execution backend.

    Notes
    -----
    All execution will happen sequentially.
    """
    from unidist.core.backends.pyseq.core import init

    init()



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

import sys
import pytest
import time

import unidist
from unidist.config import Backend, CpuCount
from unidist.core.base.common import BackendName
from .utils import assert_equal, task, TestActor

unidist.init()


@pytest.mark.skipif(
    Backend.get() == BackendName.PYMP,
    reason="Hangs on `pymp` backend. Details are in https://github.com/modin-project/unidist/issues/64.",
)
@pytest.mark.skipif(
    sys.platform == "win32" and Backend.get() == BackendName.PYMP,
    reason="Details are in https://github.com/modin-project/unidist/issues/70.",
)
@pytest.mark.parametrize(
    "source",
    ["put", "task", "actor"],
)
def test_is_object_ref(source):
    if source == "put":
        object_ref = unidist.put(1)
    elif source == "task":
        object_ref = task.remote(3)
    else:
        object_ref = TestActor.remote().task.remote(1)
    assert_equal(unidist.is_object_ref(object_ref), True)


@pytest.mark.skipif(
    Backend.get() == BackendName.PYMP,
    reason="Hangs on `pymp` backend. Details are in https://github.com/modin-project/unidist/issues/64.",
)
def test_wait():
    @unidist.remote
    def foo():
        time.sleep(3)
        return 1

    object_refs = [task.remote(5), foo.remote()]
    ready, not_ready = unidist.wait(object_refs, num_returns=1)
    assert_equal(len(ready), 1)
    assert_equal(len(not_ready), 1)
    ready, not_ready = unidist.wait(object_refs, num_returns=2)
    assert_equal(len(ready), 2)
    assert_equal(len(not_ready), 0)

    # test for https://github.com/modin-project/unidist/issues/354
    object_refs = [unidist.put(1), unidist.put(2)]
    ready, not_ready = unidist.wait(object_refs, num_returns=1)
    assert_equal(len(ready), 1)
    assert_equal(len(not_ready), 1)
    ready, not_ready = unidist.wait(object_refs, num_returns=2)
    assert_equal(len(ready), 2)
    assert_equal(len(not_ready), 0)


def test_get_ip():
    import socket

    try:
        assert_equal(unidist.get_ip(), socket.gethostbyname(socket.gethostname()))
    # Sometimes Ray returns localhost IP address in GH actions so we check this too
    except AssertionError:
        assert_equal(unidist.get_ip(), "127.0.0.1")


def test_num_cpus():
    if Backend.get() == BackendName.PYSEQ:
        assert_equal(unidist.num_cpus(), 1)
    else:
        assert_equal(unidist.num_cpus(), CpuCount.get())


def test_cluster_resources():
    assert_equal(
        unidist.cluster_resources(), {unidist.get_ip(): {"CPU": unidist.num_cpus()}}
    )



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""`SharedObjectStore` functionality."""

import os
import sys
import time
import warnings
import psutil
import weakref

from unidist.core.backends.mpi.core._memory import parallel_memcopy, fill
from unidist.core.backends.mpi.utils import ImmutableDict

try:
    import mpi4py
except ImportError:
    raise ImportError(
        "Missing dependency 'mpi4py'. Use pip or conda to install it."
    ) from None

from unidist.config.backends.mpi.envvars import (
    MpiSharedObjectStoreMemory,
    MpiSharedServiceMemory,
    MpiSharedObjectStoreThreshold,
    MpiBackoff,
)
from unidist.core.backends.mpi.core import common, communication
from unidist.core.backends.mpi.core.serialization import (
    deserialize_complex_data,
)

# TODO: Find a way to move this after all imports
mpi4py.rc(recv_mprobe=False, initialize=False)
from mpi4py import MPI  # noqa: E402


class WinLock:
    """
    Class that helps to synchronize the write to shared memory.

    Parameters
    ----------
    win : MPI.win
        The MPI window that was used to allocate the shared memory.

    Notes
    -----
    This class should be used as a context manager.
    """

    def __init__(self, win):
        self.win = win

    def __enter__(self):
        """Lock the current MPI.Window for other processes."""
        self.win.Lock(communication.MPIRank.MONITOR)

    def __exit__(self, *args):
        """Unlock the current MPI.Window for other processes."""
        self.win.Unlock(communication.MPIRank.MONITOR)


class SharedObjectStore:
    """
    Class that provides access to data in shared memory.

    Notes
    -----
    This class initializes and manages shared memory.
    """

    __instance = None

    # Service constants defining the structure of the service buffer
    # The amount of service information for one data object in shared memory.
    INFO_SIZE = 4
    # Index of service information for the first part of the DataID.
    WORKER_ID_INDEX = 0
    # Index of service information for the second part of the DataID.
    DATA_NUMBER_INDEX = 1
    # Index of service information for the first shared memory index where the data is located.
    FIRST_DATA_INDEX = 2
    # Index of service information to count the number of data references,
    # which shows how many processes are using this data.
    REFERENCES_NUMBER = 3

    def __init__(self):
        # The `MPI.Win` object to manage shared memory for data
        self.win = None
        # The `MPI.Win` object to manage shared memory for service purposes
        self.service_win = None
        # `MPI.memory` object for reading/writing data from/to shared memory
        self.shared_buffer = None
        # `memoryview` object for reading/writing data from/to service shared memory.
        # Service shared buffer includes service information about written shared data.
        # The service info is set by the worker who sends the data to shared memory
        # and is removed by the monitor if the data is cleared.
        # The service info indicates that the current data is written to shared memory
        # and shows the actual location and number of references.
        self.service_shared_buffer = None
        # Length of shared memory buffer in bytes
        self.shared_memory_size = None
        # Length of service shared memory buffer in items
        self.service_info_max_count = None

        mpi_state = communication.MPIState.get_instance()
        # Initialize all properties above
        if common.is_shared_memory_supported(raise_warning=mpi_state.is_root_process()):
            self._allocate_shared_memory()

        # Logger will be initialized after `communicator.MPIState`
        self.logger = None
        # Shared memory range {DataID: dict}
        # Shared information is the necessary information to properly deserialize data from shared memory.
        self._shared_info = weakref.WeakKeyDictionary()
        # The list of `weakref.finalize` which should be canceled before closing the shared memory.
        self.finalizers = []

    def _get_allowed_memory_size(self):
        """
        Get allowed memory size for allocate shared memory.

        Returns
        -------
        int
            The number of bytes available to allocate shared memory.
        """
        virtual_memory = psutil.virtual_memory().total
        if sys.platform.startswith("linux"):
            shm_fd = os.open("/dev/shm", os.O_RDONLY)
            try:
                shm_stats = os.fstatvfs(shm_fd)
                system_memory = shm_stats.f_bsize * shm_stats.f_bavail
                if system_memory / (virtual_memory / 2) < 0.99:
                    warnings.warn(
                        f"The size of /dev/shm is too small ({system_memory} bytes). The required size "
                        + f"at least half of RAM ({virtual_memory // 2} bytes). Please, delete files in /dev/shm or "
                        + "increase size of /dev/shm with --shm-size in Docker."
                    )
            finally:
                os.close(shm_fd)
        else:
            system_memory = virtual_memory
        return system_memory

    def _allocate_shared_memory(self):
        """
        Allocate shared memory.
        """
        mpi_state = communication.MPIState.get_instance()

        shared_object_store_memory = MpiSharedObjectStoreMemory.get()
        shared_service_memory = MpiSharedServiceMemory.get()
        # Use only 95% of available shared memory because
        # the rest is needed for intermediate shared buffers
        # handled by MPI itself for communication of small messages.
        allowed_memory_size = int(self._get_allowed_memory_size() * 0.95)

        if shared_object_store_memory is not None:
            if shared_service_memory is not None:
                self.shared_memory_size = shared_object_store_memory
                self.service_memory_size = shared_service_memory
            else:
                self.shared_memory_size = shared_object_store_memory
                # To avoid division by 0
                if MpiSharedObjectStoreThreshold.get() > 0:
                    self.service_memory_size = min(
                        # allowed memory size for service buffer
                        allowed_memory_size - self.shared_memory_size,
                        # maximum amount of memory required for the service buffer
                        (self.shared_memory_size // MpiSharedObjectStoreThreshold.get())
                        * (self.INFO_SIZE * MPI.LONG.size),
                    )
                else:
                    self.service_memory_size = (
                        allowed_memory_size - self.shared_memory_size
                    )
        else:
            if shared_service_memory is not None:
                self.service_memory_size = shared_service_memory
                self.shared_memory_size = allowed_memory_size - self.service_memory_size
            else:
                A = allowed_memory_size
                B = MpiSharedObjectStoreThreshold.get()
                C = self.INFO_SIZE * MPI.LONG.size
                # "x" is shared_memory_size
                # "y" is service_memory_size

                # requirements:
                # x + y = A
                # y = min[ (x/B) * C, 0.01 * A ]

                # calculation results:
                # if B > 99 * C:
                # x = (A * B) / (B + C)
                # y = (A * C) / (B + C)
                # else:
                # x = 0.99 * A
                # y = 0.01 * A

                if B > 99 * C:
                    self.shared_memory_size = (A * B) // (B + C)
                    self.service_memory_size = (A * C) // (B + C)
                else:
                    self.shared_memory_size = int(0.99 * A)
                    self.service_memory_size = int(0.01 * A)

        if self.shared_memory_size > allowed_memory_size:
            raise ValueError(
                "Memory for shared object storage cannot be allocated "
                + "because the value set to `MpiSharedObjectStoreMemory` exceeds the available memory."
            )

        if self.service_memory_size > allowed_memory_size:
            raise ValueError(
                "Memory for shared service storage cannot be allocated "
                + "because the value set to `MpiSharedServiceMemory` exceeds the available memory."
            )

        if self.shared_memory_size + self.service_memory_size > allowed_memory_size:
            raise ValueError(
                "The sum of the `MpiSharedObjectStoreMemory` and `MpiSharedServiceMemory` values is greater "
                + "than the available amount of memory."
            )

        # Shared memory is allocated only once by the monitor process.
        info = MPI.Info.Create()
        info.Set("alloc_shared_noncontig", "true")
        self.win = MPI.Win.Allocate_shared(
            (
                self.shared_memory_size * MPI.BYTE.size
                if mpi_state.is_monitor_process()
                else 0
            ),
            MPI.BYTE.size,
            comm=mpi_state.host_comm,
            info=info,
        )
        self.shared_buffer, _ = self.win.Shared_query(communication.MPIRank.MONITOR)

        self.service_info_max_count = (
            self.service_memory_size
            // (self.INFO_SIZE * MPI.LONG.size)
            * self.INFO_SIZE
        )
        self.service_win = MPI.Win.Allocate_shared(
            (
                self.service_info_max_count * MPI.LONG.size
                if mpi_state.is_monitor_process()
                else 0
            ),
            MPI.LONG.size,
            comm=mpi_state.host_comm,
            info=info,
        )
        service_buffer, _ = self.service_win.Shared_query(communication.MPIRank.MONITOR)
        self.service_shared_buffer = memoryview(service_buffer).cast("l")
        # Set -1 to the service buffer because 0 is a valid value and may be recognized by mistake.
        if mpi_state.is_monitor_process():
            fill(self.service_shared_buffer, -1)

    def _parse_data_id(self, data_id):
        """
        Parse `DataID` object to pair of int.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            The data identifier to be converted to a numerical form

        Returns
        -------
        tuple
            Pair of int that define the DataID
        """
        splited_id = str(data_id).replace(")", "").split("_")
        return int(splited_id[1]), int(splited_id[3])

    def _increment_ref_number(self, data_id, service_index):
        """
        Increment the number of references to indicate to the monitor that this data is being used.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
        service_index : int
            The service buffer index.

        Notes
        -----
        This function create `weakref.finalizer' with decrement function which will be called after data_id collecting.
        """
        if MPI.Is_finalized():
            return
        if service_index is None:
            raise KeyError(
                "it is not possible to increment the reference number for this data_id because it is not part of the shared data"
            )
        with WinLock(self.service_win):
            prev_ref_number = self.service_shared_buffer[
                service_index + self.REFERENCES_NUMBER
            ]
            self.service_shared_buffer[service_index + self.REFERENCES_NUMBER] = (
                prev_ref_number + 1
            )
            self.logger.debug(
                f"Rank {communication.MPIState.get_instance().global_rank}: Increment references number for {data_id} from {prev_ref_number} to {prev_ref_number + 1}"
            )
        self.finalizers.append(
            weakref.finalize(
                data_id, self._decrement_ref_number, str(data_id), service_index
            )
        )

    def _decrement_ref_number(self, data_id, service_index):
        """
        Decrement the number of references to indicate to the monitor that this data is no longer used.

        When references count is 0, it can be cleared.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.
        service_index : int
            The service buffer index.

        Notes
        -----
        This function is called in `weakref.finalizer' after data_id collecting.
        """
        # we must set service_index in args because the shared_info will be deleted before than this function is called
        if MPI.Is_finalized():
            return
        if self._check_service_info(data_id, service_index):
            with WinLock(self.service_win):
                prev_ref_number = self.service_shared_buffer[
                    service_index + self.REFERENCES_NUMBER
                ]
                self.service_shared_buffer[service_index + self.REFERENCES_NUMBER] = (
                    prev_ref_number - 1
                )
                self.logger.debug(
                    f"Rank {communication.MPIState.get_instance().global_rank}: Decrement references number for {data_id} from {prev_ref_number} to {prev_ref_number - 1}"
                )

    def _put_service_info(self, service_index, data_id, first_index):
        """
        Set service information about written shared data.

        Parameters
        ----------
        service_index : int
            The service buffer index.
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.
        first_index : int
            The first index of data in the shared buffer.

        Notes
        -----
        This information must be set after writing data to shared memory.
        """
        worker_id, data_number = self._parse_data_id(data_id)

        with WinLock(self.service_win):
            self.service_shared_buffer[service_index + self.FIRST_DATA_INDEX] = (
                first_index
            )
            self.service_shared_buffer[service_index + self.REFERENCES_NUMBER] = 1
            self.service_shared_buffer[service_index + self.DATA_NUMBER_INDEX] = (
                data_number
            )
            self.service_shared_buffer[service_index + self.WORKER_ID_INDEX] = worker_id

        self.finalizers.append(
            weakref.finalize(
                data_id, self._decrement_ref_number, str(data_id), service_index
            )
        )

    def _check_service_info(self, data_id, service_index):
        """
        Check if the `data_id` is in the shared memory on the current host.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.
        service_index : int
            The service buffer index.

        Returns
        -------
        bool
            Return the ``True`` status if `data_id` is in the shared memory.

        Notes
        -----
        This check ensures that the data is physically located in shared memory.
        """
        worker_id, data_number = self._parse_data_id(data_id)
        w_id = self.service_shared_buffer[service_index + self.WORKER_ID_INDEX]
        d_id = self.service_shared_buffer[service_index + self.DATA_NUMBER_INDEX]
        return w_id == worker_id and d_id == data_number

    def _put_shared_info(self, data_id, shared_info):
        """
        Put required information to deserialize `data_id`.

        Parameters
        ----------
        data_id : uunidist.core.backends.mpi.core.common.MpiDataID
        shared_info : unidist.core.backends.mpi.utils.ImmutableDict
            Information required for data deserialization
        """
        if not isinstance(shared_info, ImmutableDict):
            raise ValueError("Shared info should be immutable.")
        if data_id not in self._shared_info:
            self._shared_info[data_id] = shared_info

    def _read_from_shared_buffer(self, data_id, shared_info):
        """
        Read and deserialize data from the shared buffer.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
        shared_info : dict
            Information for correct deserialization.

        Returns
        -------
        object
            Data for the current Id.
        """
        buffer_lens = shared_info["raw_buffers_len"]
        buffer_count = shared_info["buffer_count"]
        s_data_len = shared_info["s_data_len"]
        service_index = shared_info["service_index"]

        first_index = self.service_shared_buffer[service_index + self.FIRST_DATA_INDEX]

        s_data_last_index = first_index + s_data_len
        s_data = (
            self.shared_buffer[first_index:s_data_last_index].cast("b").toreadonly()
        )
        prev_last_index = s_data_last_index
        raw_buffers = []
        for raw_buffer_len in buffer_lens:
            raw_last_index = prev_last_index + raw_buffer_len
            raw_buffers.append(
                self.shared_buffer[prev_last_index:raw_last_index]
                .cast("b")
                .toreadonly()
            )
            prev_last_index = raw_last_index

        data = deserialize_complex_data(s_data, raw_buffers, buffer_count)
        self.logger.debug(
            f"Rank {communication.MPIState.get_instance().global_rank}: Get {data_id} from {first_index} to {prev_last_index}. Service index: {service_index}"
        )
        return data

    def _write_to_shared_buffer(self, data_id, reservation_data, serialized_data):
        """
        Write serialized data to the shared buffer.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
        reservation_data : dict
            Information about the reserved space in shared memory for the current DataID.
        serialized_data : dict
            Serialized data.

        Returns
        -------
        dict
            Information required for correct data deserialization
        """
        first_index = reservation_data["first_index"]
        last_index = reservation_data["last_index"]
        service_index = reservation_data["service_index"]

        s_data = serialized_data["s_data"]
        raw_buffers = serialized_data["raw_buffers"]
        buffer_count = serialized_data["buffer_count"]
        s_data_len = len(s_data)

        buffer_lens = []
        s_data_first_index = first_index
        s_data_last_index = s_data_first_index + s_data_len

        if s_data_last_index > last_index:
            raise ValueError("Not enough shared space for data")
        self.shared_buffer[s_data_first_index:s_data_last_index] = s_data

        last_prev_index = s_data_last_index
        for i, raw_buffer in enumerate(raw_buffers):
            raw_buffer_first_index = last_prev_index
            raw_buffer_len = len(raw_buffer)
            last_prev_index = raw_buffer_first_index + len(raw_buffer)
            if last_prev_index > last_index:
                raise ValueError(f"Not enough shared space for {i} raw_buffer")

            parallel_memcopy(
                raw_buffer,
                self.shared_buffer[raw_buffer_first_index:last_prev_index],
                6,
            )

            buffer_lens.append(raw_buffer_len)

        self.logger.debug(
            f"Rank {communication.MPIState.get_instance().global_rank}: PUT {data_id} from {first_index} to {last_prev_index}. Service index: {service_index}"
        )

        return common.MetadataPackage.get_shared_info(
            data_id, s_data_len, buffer_lens, buffer_count, service_index
        )

    def _sync_shared_memory_from_another_host(
        self, comm, data_id, owner_rank, first_index, last_index, service_index
    ):
        """
        Receive shared data from another host including the owner's rank

        Parameters
        ----------
        comm : object
            MPI communicator object.
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            Data identifier
        owner_rank : int
            Rank of the owner process.
        first_index : int
            First index in shared memory.
        last_index : int
            Last index in shared memory.
        service_index : int
            Service buffer index.

        Notes
        -----
        After writting data, service information should be set.
        """
        sh_buf = self.get_shared_buffer(first_index, last_index)
        # recv serialized data to shared memory
        owner_monitor = (
            communication.MPIState.get_instance().get_monitor_by_worker_rank(owner_rank)
        )
        communication.send_simple_operation(
            comm,
            operation_type=common.Operation.REQUEST_SHARED_DATA,
            operation_data={"id": data_id},
            dest_rank=owner_monitor,
        )
        communication.mpi_recv_buffer(comm, owner_monitor, sh_buf)
        self.logger.debug(
            f"Rank {communication.MPIState.get_instance().global_rank}: Sync_copy {data_id} from {owner_rank} rank. Put data from {first_index} to {last_index}. Service index: {service_index}"
        )

    @classmethod
    def get_instance(cls):
        """
        Get instance of ``SharedObjectStore``.

        Returns
        -------
        SharedObjectStore
        """
        if cls.__instance is None:
            cls.__instance = SharedObjectStore()
        if cls.__instance.logger is None:
            logger_name = f"shared_store_{communication.MPIState.get_instance().host}"
            cls.__instance.logger = common.get_logger(logger_name, f"{logger_name}.log")
        return cls.__instance

    def is_allocated(self):
        """
        Check if the shared memory is allocated and ready to put data.x

        Returns
        -------
        bool
            True ot False.
        """
        return self.shared_buffer is not None

    def should_be_shared(self, data):
        """
        Check if data should be sent using shared memory.

        Parameters
        ----------
        data : dict
            Serialized data to check its size.

        Returns
        -------
        bool
            Return the ``True`` status if data should be sent using shared memory.
        """
        data_size = len(data["s_data"]) + sum([len(buf) for buf in data["raw_buffers"]])
        return data_size > MpiSharedObjectStoreThreshold.get()

    def contains(self, data_id):
        """
        Check if the store contains the `data_id` information required to deserialize the data.

        Returns
        -------
        bool
            Return the ``True`` status if shared store contains required information.

        Notes
        -----
        This check does not ensure that the data is physically located in shared memory.
        """
        return data_id in self._shared_info

    def get_shared_info(self, data_id):
        """
        Get required information to correct deserialize `data_id` from shared memory.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID

        Returns
        -------
        dict
            Information required for data deserialization
        """
        return self._shared_info[data_id]

    def get_ref_number(self, data_id, service_index):
        """
        Get current references count of data_id by service index.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
        service_index : int
            The service buffer index.

        Returns
        -------
        int
            The number of references to this data_id
        """
        # we must to set service_index in args because this function is called from monitor which can not known the shared_info
        if not self._check_service_info(data_id, service_index):
            return 0
        return self.service_shared_buffer[service_index + self.REFERENCES_NUMBER]

    def get_shared_buffer(self, first_index, last_index):
        """
        Get the requested range of shared memory

        Parameters
        ----------
        first_index : int
            Start of the requested range.
        last_index : int
            End of the requested range. (excluding)

        Notes
        -----
        This function is used to synchronize shared memory between different hosts.
        """
        return self.shared_buffer[first_index:last_index]

    def delete_service_info(self, data_id, service_index):
        """
        Delete service information for the current data Id.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
        service_index : int
            The service buffer index.

        Notes
        -----
        This function should be called by the monitor during the cleanup of shared data.
        """
        with WinLock(self.service_win):
            # Read actual value
            old_worker_id = self.service_shared_buffer[
                service_index + self.WORKER_ID_INDEX
            ]
            old_data_id = self.service_shared_buffer[
                service_index + self.DATA_NUMBER_INDEX
            ]
            old_first_index = self.service_shared_buffer[
                service_index + self.FIRST_DATA_INDEX
            ]
            old_references_number = self.service_shared_buffer[
                service_index + self.REFERENCES_NUMBER
            ]

            # check if data_id is correct
            if self._parse_data_id(data_id) == (old_worker_id, old_data_id):
                self.service_shared_buffer[service_index + self.WORKER_ID_INDEX] = -1
                self.service_shared_buffer[service_index + self.DATA_NUMBER_INDEX] = -1
                self.service_shared_buffer[service_index + self.FIRST_DATA_INDEX] = -1
                self.service_shared_buffer[service_index + self.REFERENCES_NUMBER] = -1
                self.logger.debug(
                    f"Rank {communication.MPIState.get_instance().global_rank}: Clear {data_id}. Service index: {service_index} First index: {old_first_index} References number: {old_references_number}"
                )
            else:
                self.logger.debug(
                    f"Rank {communication.MPIState.get_instance().global_rank}: Did not clear {data_id}, because there are was written another data_id: Data_ID(rank_{old_worker_id}_id_{old_data_id})"
                )
                self.logger.debug(
                    f"Service index: {service_index} First index: {old_first_index} References number: {old_references_number}"
                )
                raise RuntimeError("Unexpected data_id for cleanup shared memory")

    def put(self, data_id, serialized_data):
        """
        Put data into shared memory.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
        serialized_data : dict
            Serialized data to put into the storage.
        """
        mpi_state = communication.MPIState.get_instance()

        data_size = len(serialized_data["s_data"]) + sum(
            [len(buf) for buf in serialized_data["raw_buffers"]]
        )
        # reserve shared memory
        reservation_data = communication.send_reserve_operation(
            mpi_state.global_comm, data_id, data_size
        )
        service_index = reservation_data["service_index"]
        first_index = reservation_data["first_index"]

        # write into shared buffer
        shared_info = self._write_to_shared_buffer(
            data_id, reservation_data, serialized_data
        )

        # put service info
        self._put_service_info(service_index, data_id, first_index)

        # put shared info
        self._put_shared_info(data_id, shared_info)

    def get(self, data_id, owner_rank=None, shared_info=None):
        """
        Get data from another worker using shared memory.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.
        owner_rank : int, default: None
            The rank that sent the data.
            This value is used to synchronize data in shared memory between different hosts
            if the value is not ``None``.
        shared_info : dict, default: None
            The necessary information to properly deserialize data from shared memory.
            If `shared_info` is ``None``, the data already exists in shared memory in the current process.
        """
        if shared_info is None:
            shared_info = self.get_shared_info(data_id)
        else:
            mpi_state = communication.MPIState.get_instance()
            s_data_len = shared_info["s_data_len"]
            raw_buffers_len = shared_info["raw_buffers_len"]
            service_index = shared_info["service_index"]
            buffer_count = shared_info["buffer_count"]

            # check data in shared memory
            if not self._check_service_info(data_id, service_index):
                # reserve shared memory
                shared_data_len = s_data_len + sum([buf for buf in raw_buffers_len])
                reservation_info = communication.send_reserve_operation(
                    mpi_state.global_comm, data_id, shared_data_len
                )

                service_index = reservation_info["service_index"]
                # check if worker should sync shared buffer or it is doing by another worker
                if reservation_info["is_first_request"]:
                    # syncronize shared buffer
                    if owner_rank is None:
                        raise ValueError(
                            "The data is not in the host's shared memory and the data must be synchronized, "
                            + "but the owner rank is not defined."
                        )

                    self._sync_shared_memory_from_another_host(
                        mpi_state.global_comm,
                        data_id,
                        owner_rank,
                        reservation_info["first_index"],
                        reservation_info["last_index"],
                        service_index,
                    )
                    # put service info
                    self._put_service_info(
                        service_index, data_id, reservation_info["first_index"]
                    )
                else:
                    # wait while another worker syncronize shared buffer
                    while not self._check_service_info(data_id, service_index):
                        time.sleep(MpiBackoff.get())

            # put shared info with updated data_id and service_index
            shared_info = common.MetadataPackage.get_shared_info(
                data_id, s_data_len, raw_buffers_len, buffer_count, service_index
            )
            self._put_shared_info(data_id, shared_info)

            # increment ref
            self._increment_ref_number(data_id, shared_info["service_index"])

        # read from shared buffer and deserialized
        return self._read_from_shared_buffer(data_id, shared_info)

    def finalize(self):
        """
        Release used resources.

        Notes
        -----
        Shared store should be finalized before MPI.Finalize().
        """
        if self.win is not None:
            self.win.Free()
            self.win = None
        if self.service_win is not None:
            self.service_win.Free()
            self.service_win = None
        for f in self.finalizers:
            f.detach()



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""``DataID`` class functionality."""

import uuid


class DataID:
    """Class that holds unique identifier.

    Parameters
    ----------
    id_value : object
        Any comparable and hashable ID value.
    """

    def __init__(self, id_value=None):
        self._id = id_value if id_value is not None else uuid.uuid4().hex

    def __repr__(self):
        return f"DataID({self._id})"

    def __eq__(self, other):
        return hasattr(other, "_id") and self._id == other._id

    def __hash__(self):
        return hash(self._id)


def is_data_id(arg):
    """Check if argument is an instance of ``DataID``.

    Parameters
    ----------
    arg : object
        Object to check.

    Returns
    -------
    bool
        ``True`` if argument is an instance of ``DataID``.
    """
    return isinstance(arg, DataID)



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Python Sequential backend core functionality."""

from .api import put, get, submit, init, is_initialized

__all__ = ["put", "get", "submit", "init", "is_initialized"]



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Workers related functionality."""

import cloudpickle as pkl
from multiprocessing import (
    Process,
    JoinableQueue,
)

from unidist.config import CpuCount
from unidist.core.backends.common.data_id import DataID
from unidist.core.backends.pymp.core.object_store import ObjectStore


class Worker(Process):
    """
    Class-process that executes tasks from `self.task_queue`.

    Parameters
    ----------
    task_queue : multiprocessing.JoinableQueue
        A queue of task to execute.
    obj_store : unidist.core.backends.pymp.core.object_store.ObjectStore
        Shared object storage to read/write data.
    """

    def __init__(self, task_queue, obj_store):
        Process.__init__(self, daemon=True)
        self.task_queue = task_queue
        self._obj_store = obj_store

    def run(self):
        """Run main infinite loop of process to execute tasks from `self.task_queue`."""
        while 1:
            task = self.task_queue.get()
            task = pkl.loads(task)

            data_ids = task.data_ids
            try:
                value = task()
            except Exception as e:
                if isinstance(data_ids, list) and len(data_ids) > 1:
                    for i, data_id in enumerate(data_ids):
                        self._obj_store.store_delayed[data_id] = e
                else:
                    self._obj_store.store_delayed[data_ids] = e
            else:
                if data_ids is not None:
                    if isinstance(data_ids, list) and len(data_ids) > 1:
                        for data_id, val in zip(data_ids, value):
                            self._obj_store.store_delayed[data_id] = val
                    else:
                        self._obj_store.store_delayed[data_ids] = value
            finally:
                self.task_queue.task_done()
        return

    def add_task(self, task):
        """
        Add `task` to `self.task_queue`.

        Parameters
        ----------
        task : unidist.core.backends.pymp.core.process_manager.Task
            Task to be added in the queue.
        """
        self.task_queue.put(task)


class ProcessManager:
    """
    Class that controls worker pool and assings task to workers.

    Parameters
    ----------
    num_workers : int, optional
        Number of worker-processes to start. If isn't provided,
        will be equal to number of CPUs.

    Notes
    -----
    Constructor starts `num_workers` Multiprocessing Workers.
    """

    __instance = None

    def __init__(self, num_workers=None):
        if ProcessManager.__instance is None:
            if num_workers is None:
                num_workers = CpuCount.get()
            self.workers = [None] * num_workers
            self.grabbed_workers = [None] * num_workers
            self.__class__._worker_id = 0

            obj_store = ObjectStore.get_instance()
            for idx in range(num_workers):
                self.workers[idx] = Worker(JoinableQueue(), obj_store)
                self.workers[idx].start()
                self.grabbed_workers[idx] = False

    @classmethod
    def get_instance(cls, num_workers=None):
        """
        Get instance of ``ProcessManager``.

        Returns
        -------
        unidist.core.backends.pymp.core.process_manager.ProcessManager
        """
        if cls.__instance is None:
            cls.__instance = ProcessManager(num_workers=num_workers)
        return cls.__instance

    def _next(self):
        """
        Get current worker index and move to another with incrementing by one.

        Returns
        -------
        int
        """
        idx = self.__class__._worker_id
        self.__class__._worker_id += 1
        if self.__class__._worker_id == len(self.workers):
            self.__class__._worker_id = 0
        return idx

    def grab_worker(self):
        """
        Grab a worker from worker pool.

        Grabbed worker is marked as `blocked` and doesn't participate
        in the tasks submission.

        Returns
        -------
        unidist.core.backends.pymp.core.process_manager.Worker
            Grabbed worker.
        int
            Index of grabbed worker.
        """
        for idx, is_grabbed in enumerate(self.grabbed_workers):
            if not is_grabbed:
                self.grabbed_workers[idx] = True
                return self.workers[idx], idx
        raise RuntimeError("Actor can`t be run, no available workers.")

    def free_worker(self, idx):
        """
        Free worker by index `idx`.

        Parameters
        ----------
        idx : int
            Index of worker to be freed.
        """
        self.grabbed_workers[idx] = False

    def submit(self, task):
        """
        Add `task` to task queue of one of workers using round-robin.

        Parameters
        ----------
        task : unidist.core.backends.pymp.core.process_manager.Task
            Task to be added in task queue.
        """
        num_skipped = 0

        while num_skipped < len(self.workers):
            idx = self._next()
            if not self.grabbed_workers[idx]:
                self.workers[idx].add_task(task)
                return
            else:
                num_skipped += 1

        raise RuntimeError("Task can`t be run, no available workers.")


class Task:
    """
    Class poses as unified callable object to execute in Multiprocessing Worker.

    Parameters
    ----------
    func : callable
        A function to be called in object invocation.
    data_ids : unidist.core.backends.common.data_id.DataID or list
        ``DataID``-(s) associated with result(s) of `func` invocation.
    obj_store : unidist.core.backends.pymp.core.object_store.ObjectStore
        Object storage to share data between workers.
    *args : iterable
        Positional arguments to be passed in the `func`.
    **kwargs : dict
        Keyword arguments to be passed in the `func`.
    """

    def __init__(self, func, data_ids, obj_store, *args, **kwargs):
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self.data_ids = data_ids
        self.obj_store = obj_store

    def __call__(self):
        """
        Execute `self._func`.

        If `self._args`/`self._kwargs` has ``DataID`` objects,
        automaterialize happens.

        Returns
        -------
        object
            The result of `self._func` invocation.
        """
        materialized_args = [
            self.obj_store.get(arg) if isinstance(arg, DataID) else arg
            for arg in self._args
        ]
        materialized_kwargs = {
            key: self.obj_store.get(value) if isinstance(value, DataID) else value
            for key, value in self._kwargs.items()
        }
        return self._func(*materialized_args, **materialized_kwargs)
